"""This module tests classes related to LabelSchema"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
from networkx.classes.reportviews import EdgeDataView, NodeView, OutMultiEdgeDataView

from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import (
    LabelEntity,
    LabelGraph,
    LabelGroup,
    LabelGroupDoesNotExistException,
    LabelGroupExistsException,
    LabelGroupType,
    LabelSchemaEntity,
    LabelTree,
    ScoredLabel,
)
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestLabelSchema:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree(self, label_schema_example):
        """
        <b>Description:</b>
        Check that childs and parents of Labels can be found correctly

        <b>Input data:</b>
        A simple LabelTree

        <b>Expected results:</b>
        Test passes if the correct parent and child labels can be found

        <b>Steps</b>
        1. Create LabelTree
        2. Find parents of Labels
        3. Find children of Label
        """
        threat = label_schema_example.new_label_by_name("threat")
        benign = label_schema_example.new_label_by_name("benign")
        gun = label_schema_example.new_label_by_name("gun")
        rifle = label_schema_example.new_label_by_name("rifle")
        bottle = label_schema_example.new_label_by_name("bottle")

        label_tree = LabelTree()
        label_tree.add_child(threat, gun)
        label_tree.add_child(threat, rifle)
        label_tree.add_child(benign, bottle)

        assert label_tree.get_parent(gun) == threat
        assert label_tree.get_parent(rifle) == threat
        assert label_tree.get_parent(bottle) == benign
        assert label_tree.get_parent(threat) is None

        threat_children = label_tree.get_children(threat)
        assert rifle in threat_children

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_exclusive(self, label_schema_example):
        """
        <b>Description:</b>
        Tests that the are_exclusive method properly computes the labels that are exclusive with each other

        <b>Input data:</b>
        Exclusive group: flowering,vegetative,no_plant
        Exclusive group: bee

        <b>Expected results:</b>
        Test passes if the exclusive groups are correctly saved and retrieved

        <b>Steps</b>
        1. Create LabelRelation with exclusive group
        2. Check that relations are exclusive
        3. Check that number of relations is correct
        4. Add new exclusive group
        5. Check that exclusive relations still resolve correctly
        """
        flowering = label_schema_example.flowering
        no_plant = label_schema_example.no_plant
        vegetative = label_schema_example.vegetative

        label_schema = LabelSchemaEntity()
        label_schema.add_group(
            LabelGroup(
                "plant_state",
                [flowering, no_plant, vegetative],
            )
        )

        assert label_schema.are_exclusive(flowering, no_plant)
        assert label_schema.are_exclusive(vegetative, no_plant)
        assert label_schema.are_exclusive(vegetative, flowering)

        bee = label_schema_example.new_label_by_name("bee")
        label_schema.add_group(LabelGroup("bee_state", [bee], LabelGroupType.EXCLUSIVE))

        assert not label_schema.are_exclusive(flowering, bee)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_complex(self, label_schema_example):
        """
        <b>Description:</b>
        Tests that exclusivity is computed correctly in a schema with complex hierarchy

        <b>Input data:</b>
        Exclusive groups: plant, animal
        Exclusive groups: tree, bush (children of plant), mammal, insect (children of animal)
        Empty label

        <b>Expected results:</b>
        Test passes if the are_exclusive method properly computes the exclusivity between labels with various relations

        <b>Steps</b>
        1. Create LabelSchemaEntity with hierarchy
        2. Check that labels are not exclusive with their parent
        3. Check that labels are exclusive with:
         - labels in the same group
         - children of labels in the same group
         - labels in the same group as their parent
         - children of labels in the same group as their parent
        4. Add an empty label and check that it's exclusive with all other labels
        5. Make the empty label the child of a different label, and check that the empty label is not exclusive with
        its parent.
        6. Check that LabelSchema.are_exclusive is symmetric for all cases

        """
        label_schema = LabelSchemaEntity()

        plant = label_schema_example.new_label_by_name("plant")
        animal = label_schema_example.new_label_by_name("animal")
        label_schema.add_group(LabelGroup("organism_type", [plant, animal], LabelGroupType.EXCLUSIVE))

        tree = label_schema_example.new_label_by_name("tree")
        bush = label_schema_example.new_label_by_name("bush")
        label_schema.add_group(LabelGroup("plant_type", [tree, bush], LabelGroupType.EXCLUSIVE))

        label_schema.add_child(parent=plant, child=tree)
        label_schema.add_child(parent=plant, child=bush)

        insect = label_schema_example.new_label_by_name("insect")
        mammal = label_schema_example.new_label_by_name("mammal")
        label_schema.add_group(LabelGroup("animal_type", [insect, mammal], LabelGroupType.EXCLUSIVE))
        label_schema.add_child(parent=animal, child=insect)
        label_schema.add_child(parent=animal, child=mammal)

        assert not label_schema.are_exclusive(plant, tree)
        assert label_schema.are_exclusive(tree, bush)
        assert label_schema.are_exclusive(plant, insect)
        assert label_schema.are_exclusive(tree, animal)
        assert label_schema.are_exclusive(tree, insect)

        # Check that the empty label is exclusive with all labels
        empty_label = label_schema_example.new_label_by_name("empty_label", is_empty=True)
        label_schema.add_group(LabelGroup("empty_label", [empty_label], LabelGroupType.EMPTY_LABEL))
        for label_iter in label_schema.get_labels(include_empty=False):
            assert label_schema.are_exclusive(empty_label, label_iter)

        # Check that the empty label is not exclusive with its parent
        label_schema.add_child(parent=tree, child=empty_label)
        assert not label_schema.are_exclusive(tree, empty_label)

        # Check that label_schema.are_exclusive is symmetric for all cases
        for label_1 in label_schema.get_labels(include_empty=True):
            for label_2 in label_schema.get_labels(include_empty=True):
                assert label_schema.are_exclusive(label_1, label_2) == label_schema.are_exclusive(label_2, label_1)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_add_label_to_group(self, label_schema_example):
        """
        <b>Description:</b>
        Check that new labels can be added to groups

        <b>Input data:</b>
        Empty Label Tree

        <b>Expected results:</b>
        Test passes if an exclusive group can be added to the LabelTree
        as long as a group with the same name does not exist and new labels can be added to a group.

        <b>Steps</b>
        1. Create LabelRelation
        2. Create new label and add to exclusive group
        3. Create new label and attemp to create new group with the same name
        4. Add label to group by group name
        """
        label_schema = LabelSchemaEntity()
        bee = label_schema_example.new_label_by_name("bee")  # indicates presence/absence of bee
        bee_state = LabelGroup("bee_state", [bee], LabelGroupType.EXCLUSIVE)
        label_schema.add_group(bee_state)

        # Try to add an extra bee property as a new exclusive label group, but with the same name
        # as an already existing label group
        flying = label_schema_example.new_label_by_name("flying")
        with pytest.raises(ValueError):
            label_schema.add_group(LabelGroup("bee_state", [flying], LabelGroupType.EXCLUSIVE))

        label_schema.add_labels_to_group_by_group_name("bee_state", [flying])

        assert "bee_state" == bee_state.name
        assert 2 == len(bee_state.labels)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_exclusivity_with_group_addition(self, label_schema_example):
        """
        <b>Description:</b>
        Check that exclusive groups work correctly

        <b>Input data:</b>
        Exclusive Group "Vegetation"
        LabelSchemaExample instance

        <b>Expected results:</b>
        Test passes if a single label and multiple labels can be added to the exclusive group
        and the correct labels can be retrieved

        <b>Steps</b>
        1. Create LabelSchemaEntity with "Vegetation" exclusive group
        2. Add "no_plant" to exclusive group and check that "flowering" is exclusive to "no_plant"
        3. Add "vegetative" to exclusive group and check that "vegetative" is exclusive to "no_plant"
        4. Add "flowering" to exclusive group and check that "no_plant" and "vegetative" are exclusive to "flowering"
        5. Create new LabelRelation instace
        6. Add "no_plant" and "vegetative" to exclusive group "flowering"
        7. Check that "flowering" and "vegetative" are exclusive to "no_plants"
        8. Check that no_plant" and "vegetative" are exclusive to "flowering
        """
        label_schema = LabelSchemaEntity()
        label_group_name = "Vegetation"
        label_schema.add_group(
            LabelGroup(
                label_group_name,
                [label_schema_example.flowering],
                LabelGroupType.EXCLUSIVE,
            )
        )

        label_schema.add_labels_to_group_by_group_name(label_group_name, [label_schema_example.no_plant])
        exclusive_to_no_plants = label_schema.get_labels_exclusive_to(label_schema_example.no_plant)
        assert label_schema_example.flowering in exclusive_to_no_plants

        label_schema.add_labels_to_group_by_group_name(label_group_name, [label_schema_example.vegetative])
        exclusive_to_no_plants = label_schema.get_labels_exclusive_to(label_schema_example.no_plant)
        assert label_schema_example.vegetative in exclusive_to_no_plants

        exclusive_to_flowering = label_schema.get_labels_exclusive_to(label_schema_example.flowering)
        assert label_schema_example.no_plant in exclusive_to_flowering
        assert label_schema_example.vegetative in exclusive_to_flowering

        # new label schema (test adding multiple labels at once)
        label_schema = LabelSchemaEntity()
        label_group_name = "Vegetation"
        label_schema.add_group(
            LabelGroup(
                label_group_name,
                [label_schema_example.flowering],
                LabelGroupType.EXCLUSIVE,
            )
        )
        label_schema.add_labels_to_group_by_group_name(
            label_group_name,
            [label_schema_example.no_plant, label_schema_example.vegetative],
        )

        exclusive_to_no_plants = label_schema.get_labels_exclusive_to(label_schema_example.no_plant)
        assert label_schema_example.flowering in exclusive_to_no_plants
        assert label_schema_example.vegetative in exclusive_to_no_plants

        exclusive_to_flowering = label_schema.get_labels_exclusive_to(label_schema_example.flowering)
        assert label_schema_example.no_plant in exclusive_to_flowering
        assert label_schema_example.vegetative in exclusive_to_flowering

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_labelschema_equality(self, label_schema_example):
        """
        <b>Description:</b>
        Check that LabelSchemaEntity equality works correctly

        <b>Input data:</b>
        LabelSchemaEntity  instances

        <b>Expected results:</b>
        == and != operations work correctly for various inputs

        <b>Steps</b>
        1. Test LabelSchemaEntity equality
        """
        label_schema = LabelSchemaEntity()

        label_group_name = "Vegetation"
        label_schema.add_group(
            LabelGroup(
                label_group_name,
                [label_schema_example.flowering],
                LabelGroupType.EXCLUSIVE,
            )
        )

        label_schema.add_labels_to_group_by_group_name(label_group_name, [label_schema_example.no_plant])

        copy_schema = label_schema
        assert label_schema == copy_schema

        new_schema = LabelSchemaEntity()

        label_group_name = "Vegetation"
        new_schema.add_group(
            LabelGroup(
                label_group_name,
                [label_schema_example.flowering],
                LabelGroupType.EXCLUSIVE,
            )
        )

        new_schema.add_labels_to_group_by_group_name(label_group_name, [label_schema_example.vegetative])

        assert new_schema != label_schema


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestLabelGroupType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_group_type(self):
        """
        <b>Description:</b>
        Check LabelGroupType Enum class elements

        <b>Expected results:</b>
        Test passes if LabelGroupType Enum class length is equal to expected value and its elements have expected
        sequence numbers
        """
        assert len(LabelGroupType) == 2
        assert LabelGroupType.EXCLUSIVE.value == 1
        assert LabelGroupType.EMPTY_LABEL.value == 2


class Labels:
    def __init__(self):
        self.label_0 = LabelEntity(
            name="Label 0",
            domain=Domain.CLASSIFICATION,
            id=ID("0"),
            color=Color(25, 200, 166),
        )
        self.label_0_1 = LabelEntity(
            name="Label 0_1",
            domain=Domain.DETECTION,
            id=ID("0_1"),
            color=Color(40, 100, 17),
        )
        self.label_0_2 = LabelEntity(
            name="Label 0_2",
            domain=Domain.SEGMENTATION,
            id=ID("0_2"),
            color=Color(30, 80, 40),
        )
        self.label_0_1_3 = LabelEntity(
            name="Label_0_1_3",
            domain=Domain.SEGMENTATION,
            id=ID("0_1_3"),
            color=Color(40, 100, 17),
        )
        self.label_0_2_4 = LabelEntity(
            name="Label_0_2_4",
            domain=Domain.SEGMENTATION,
            id=ID("0_2_4"),
            color=Color(30, 80, 40),
        )
        self.label_0_2_5 = LabelEntity(
            name="Label_0_2_5",
            domain=Domain.SEGMENTATION,
            id=ID("0_2_5"),
            color=Color(30, 80, 40),
        )
        self.no_id_label = LabelEntity(name="No ID Label", domain=Domain.SEGMENTATION)
        self.non_included_label = LabelEntity(name="Label non included to group", domain=Domain.SEGMENTATION)


labels = Labels()


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestLabelGroup:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_group_initialization(self):
        """
        <b>Description:</b>
        Check LabelGroup class object initialization

        <b>Input data:</b>
        LabelGroup object with specified name, labels, group type and ID parameters

        <b>Expected results:</b>
        Test passes if LabelGroup object id, labels, name and group_type attributes return expected values

        <b>Steps</b>
        1. Check id, labels, name and group_type attributes of LabelGroup object with not specified group_type
        parameter
        2. Check id, labels, name and group_type attributes of LabelGroup object with not specified id parameter
        """
        # Checking attributes of LabelGroup object with specified id and not specified group_type parameters
        no_group_type_label_group = LabelGroup(
            name="Type non-specified specified label group",
            labels=[labels.label_0_1, labels.no_id_label, labels.label_0],
            id=ID("1"),
        )
        assert no_group_type_label_group.id_ == "1"
        # Expected ascending sorting of labels
        assert no_group_type_label_group.labels == [
            labels.no_id_label,
            labels.label_0,
            labels.label_0_1,
        ]
        assert no_group_type_label_group.name == "Type non-specified specified label group"
        assert no_group_type_label_group.group_type == LabelGroupType.EXCLUSIVE
        assert isinstance(no_group_type_label_group.minimum_label_id, ID)
        assert no_group_type_label_group.minimum_label_id == ""
        # Checking attributes of LabelGroup object with specified group_type and not specified id parameters
        no_id_label_group = LabelGroup(
            name="ID non-specified Label Group",
            labels=[labels.label_0_1, labels.label_0],
            group_type=LabelGroupType.EMPTY_LABEL,
        )
        # Expected randomly generated ID object with 24 characters as "id" attribute
        assert isinstance(no_id_label_group.id_, ID)
        assert len(no_id_label_group.id_) == 24
        # Expected ascending sorting of labels
        assert no_id_label_group.labels == [labels.label_0, labels.label_0_1]
        assert no_id_label_group.name == "ID non-specified Label Group"
        assert no_id_label_group.group_type == LabelGroupType.EMPTY_LABEL
        assert isinstance(no_id_label_group.minimum_label_id, ID)
        assert no_id_label_group.minimum_label_id == "0"

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_group_remove_label(self):
        """
        <b>Description:</b>
        Check remove_label method of LabelGroup class object

        <b>Input data:</b>
        LabelGroup objects with specified name, labels, group type and id_parameters parameters

        <b>Expected results:</b>
        Test passes if after using remove_label method values of "labels" property, "minimum_label_id" and
        "is_single_label" methods are equal to expected
        """
        label_group = LabelGroup(name="Test Label Group", labels=[labels.label_0, labels.label_0_1])
        assert not label_group.is_single_label()
        # Removing first label in "labels" property and checking values of "labels", "minimum_label_id" and
        # "is_single_label"
        label_group.remove_label(labels.label_0)
        assert label_group.labels == [labels.label_0_1]
        assert label_group.minimum_label_id == "0_1"
        assert label_group.is_single_label()
        # Removing label that not included to LabelGroup object and repeat checks
        label_group.remove_label(labels.non_included_label)
        assert label_group.labels == [labels.label_0_1]
        assert label_group.minimum_label_id == "0_1"
        assert label_group.is_single_label()

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_group_magic_methods(self):
        """
        <b>Description:</b>
        Check __eq__ and __repr__ methods of LabelGroup class object

        <b>Input data:</b>
        LabelGroup class object with specified name, labels, group type and ID parameters

        <b>Expected results:</b>
        Test passes if __eq__ and __repr__ methods return expected values

        <b>Steps</b>
        1. Check __eq__ method for LabelGroup class objects
        2. Check __repr__ method for LabelGroup class objects
        """
        group_labels = [labels.label_0, labels.label_0_1]
        name = "Test Label Group"
        group_id = ID("1")
        label_group = LabelGroup(name=name, labels=group_labels, id=group_id)
        equal_label_group = LabelGroup(name=name, labels=group_labels, id=group_id)
        no_id_specified_label_group = LabelGroup(name=name, labels=group_labels)
        # Checking __eq__ method for equal LabelGroup objects
        assert label_group == equal_label_group
        # Checking equality of LabelGroups with different "name" attributes
        assert label_group == LabelGroup(name="Different name LabelGroup", labels=group_labels, id=group_id)
        # Checking inequality of LabelGroups with different "id" attributes
        assert not label_group == no_id_specified_label_group
        # Checking inequality of LabelGroups with different "labels" attributes
        assert not label_group == LabelGroup(name=name, labels=[labels.label_0], id=group_id)
        # Checking inequality of LabelGroups with different "group_type" attributes
        assert not label_group == LabelGroup(
            name=name,
            labels=group_labels,
            id=group_id,
            group_type=LabelGroupType.EMPTY_LABEL,
        )
        # Checking inequality of LabelGroups with different object of different type
        assert not label_group == str
        # Checking __repr__ method for LabelGroup object with specified id
        assert repr(label_group) == (
            "LabelGroup(id=1, name=Test Label Group, group_type=LabelGroupType.EXCLUSIVE, " f"labels={group_labels})"
        )
        # Checking __repr__ method for LabelGroup object with [] labels and not specified id
        no_labels_no_id_label_group = LabelGroup(
            name="Null labels, random id group",
            labels=[],
            group_type=LabelGroupType.EMPTY_LABEL,
        )
        generated_id = no_labels_no_id_label_group.id_
        assert repr(no_labels_no_id_label_group) == (
            f"LabelGroup(id={generated_id}, name=Null labels, random id group,"
            f" group_type=LabelGroupType.EMPTY_LABEL, labels=[])"
        )


class CommonGraphMethods:
    @staticmethod
    def check_graph_non_list_attributes(expected_attributes_dicts: list) -> None:
        for expected_attribute_dict in expected_attributes_dicts:
            assert expected_attribute_dict.get("attribute") == expected_attribute_dict.get("expected_value")

    @staticmethod
    def check_graph_list_attributes(actual_expected_attributes_dicts: list) -> None:
        for expected_attribute_dict in actual_expected_attributes_dicts:
            attribute = expected_attribute_dict.get("attribute")
            assert isinstance(attribute, expected_attribute_dict.get("expected_type"))
            assert list(attribute) == expected_attribute_dict.get("expected_value")


class Edges:
    def __init__(self):
        self.edge_0_to_0_1 = (labels.label_0, labels.label_0_1)
        self.edge_0_to_0_2 = (labels.label_0, labels.label_0_2)
        self.edge_0_1_to_0_2 = (labels.label_0_1, labels.label_0_2)
        self.edge_0_2_to_0 = (labels.label_0_2, labels.label_0)
        self.edge_0_1_to_0_1_3 = (labels.label_0_1, labels.label_0_1_3)
        self.edge_0_2_to_0_2_4 = (labels.label_0_2, labels.label_0_2_4)
        self.edge_0_2_to_0_2_5 = (labels.label_0_2, labels.label_0_2_5)
        self.edge_0_2_4_to_0_2_5 = (labels.label_0_2_4, labels.label_0_2_5)


edges = Edges()


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestLabelGraph:
    @staticmethod
    def non_directed_label_graph() -> LabelGraph:
        non_directed_graph = LabelGraph(directed=False)
        non_directed_graph.add_edges([edges.edge_0_to_0_1, edges.edge_0_1_to_0_2])
        return non_directed_graph

    @staticmethod
    def directed_label_graph() -> LabelGraph:
        directed_graph = LabelGraph(directed=True)
        directed_graph.add_edges([edges.edge_0_to_0_1, edges.edge_0_1_to_0_2, edges.edge_0_2_to_0])
        return directed_graph

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_graph_initialization(self):
        """
        <b>Description:</b>
        Check LabelGraph class object initialization

        <b>Input data:</b>
        LabelGraph objects with specified "directed" parameter and added edges

        <b>Expected results:</b>
        Test passes if LabelGraph object "directed" attribute and "num_labels" and "type" properties are equal to
        expected

        <b>Steps</b>
        1. Check "directed" attribute and "num_labels" and "type" properties of non-directed LabelGraph object
        2. Check "directed" attribute and "num_labels" and "type" properties of directed LabelGraph object
        """
        # Checks for non-directed LabelGraph
        non_directed_graph = LabelGraph(directed=False)
        # Check for initiated non-directed LabelGraph
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": non_directed_graph.directed, "expected_value": False},
                {"attribute": non_directed_graph.num_labels, "expected_value": 0},
                {"attribute": non_directed_graph.num_nodes(), "expected_value": 0},
                {"attribute": non_directed_graph.type, "expected_value": "graph"},
            ]
        )
        # Check for non-directed LabelGraph with added edges and nodes
        non_directed_graph.add_edges([(labels.label_0, labels.label_0_1), (labels.label_0_1, labels.label_0_2)])
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": non_directed_graph.directed, "expected_value": False},
                {"attribute": non_directed_graph.num_labels, "expected_value": 3},
                {"attribute": non_directed_graph.num_nodes(), "expected_value": 3},
                {"attribute": non_directed_graph.type, "expected_value": "graph"},
            ]
        )
        expected_nodes = [labels.label_0, labels.label_0_1, labels.label_0_2]
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": non_directed_graph.edges,
                    "expected_type": EdgeDataView,
                    "expected_value": [
                        (labels.label_0, labels.label_0_1, {}),
                        (labels.label_0_1, labels.label_0_2, {}),
                    ],
                },
                {
                    "attribute": non_directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": expected_nodes,
                },
            ]
        )
        # Checks for directed LabelGraph
        directed_graph = LabelGraph(directed=True)
        # Check for initiated directed LabelGraph
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": directed_graph.directed, "expected_value": True},
                {"attribute": directed_graph.num_labels, "expected_value": 0},
                {"attribute": directed_graph.num_nodes(), "expected_value": 0},
                {"attribute": directed_graph.type, "expected_value": "graph"},
            ]
        )
        # Check for directed LabelGraph with added edges and nodes
        directed_graph.add_edges([edges.edge_0_to_0_1, edges.edge_0_1_to_0_2, edges.edge_0_2_to_0])
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": directed_graph.directed, "expected_value": True},
                {"attribute": directed_graph.num_labels, "expected_value": 3},
                {"attribute": directed_graph.num_nodes(), "expected_value": 3},
                {"attribute": directed_graph.type, "expected_value": "graph"},
            ]
        )
        expected_nodes = [labels.label_0, labels.label_0_1, labels.label_0_2]
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": directed_graph.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        (labels.label_0, labels.label_0_1, 0, {}),
                        (labels.label_0_1, labels.label_0_2, 0, {}),
                        (labels.label_0_2, labels.label_0, 0, {}),
                    ],
                },
                {
                    "attribute": directed_graph.nodes,
                    "expected_type": NodeView,
                    "expected_value": expected_nodes,
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_graph_subgraph(self):
        """
        <b>Description:</b>
        Check "subgraph" method of LabelGraph class object

        <b>Input data:</b>
        LabelGraph objects with specified directed parameter and added edges

        <b>Expected results:</b>
        Test passes if subgraph method returns expected LabelGraph instance for LabelGraph object

        <b>Steps</b>
        1. Check LabelGraph instance returned by subgraph method for non-directed LabelGraph object
        2. Check LabelGraph instance returned by subgraph method for directed LabelGraph object
        """
        # Checks for non-directed LabelGraph
        non_directed_graph = self.non_directed_label_graph()
        # Checking subgraph with one node not included to parent graph
        non_directed_subgraph = non_directed_graph.subgraph(
            [labels.label_0_1, labels.label_0_2, labels.non_included_label]
        )
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": non_directed_subgraph.directed, "expected_value": False},
                {"attribute": non_directed_subgraph.num_labels, "expected_value": 2},
                {"attribute": non_directed_subgraph.num_nodes(), "expected_value": 2},
                {"attribute": non_directed_subgraph.type, "expected_value": "graph"},
            ]
        )
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": non_directed_subgraph.edges,
                    "expected_type": EdgeDataView,
                    "expected_value": [(labels.label_0_1, labels.label_0_2, {})],
                },
                {
                    "attribute": non_directed_subgraph.nodes,
                    "expected_type": NodeView,
                    "expected_value": [labels.label_0_1, labels.label_0_2],
                },
            ]
        )

        # Checks for directed LabelGraph
        directed_graph = self.directed_label_graph()
        # Checking subgraph with one node not included to parent graph
        directed_subgraph = directed_graph.subgraph([labels.label_0_1, labels.label_0_2, labels.non_included_label])
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": directed_subgraph.directed, "expected_value": True},
                {"attribute": directed_subgraph.num_labels, "expected_value": 2},
                {"attribute": directed_subgraph.num_nodes(), "expected_value": 2},
                {"attribute": directed_subgraph.type, "expected_value": "graph"},
            ]
        )
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": directed_subgraph.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [(labels.label_0_1, labels.label_0_2, 0, {})],
                },
                {
                    "attribute": directed_subgraph.nodes,
                    "expected_type": NodeView,
                    "expected_value": [labels.label_0_1, labels.label_0_2],
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_graph_eq(self):
        """
        <b>Description:</b>
        Check __eq__ method of LabelGraph class object

        <b>Input data:</b>
        LabelGraph objects with specified directed parameter and added edges

        <b>Expected results:</b>
        Test passes if __eq__ method returns expected values

        <b>Steps</b>
        1. Check __eq__ method for non-directed LabelGraph object
        2. Check __eq__ method for directed LabelGraph object
        """
        # Checks for non-directed LabelGraph
        # Checking __eq__ method for equal non-directed LabelGraph
        non_directed_graph = self.non_directed_label_graph()
        equal_non_directed_graph = self.non_directed_label_graph()
        assert non_directed_graph == equal_non_directed_graph
        # Checking __eq__ method for equal non-directed and directed LabelGraph objects
        directed_graph = LabelGraph(directed=True)
        directed_graph.add_edges([edges.edge_0_to_0_1, edges.edge_0_1_to_0_2])
        assert non_directed_graph != directed_graph
        # Checking __eq__ method for non-directed LabelGraph objects with different edges
        non_directed_graph_different_edges = self.non_directed_label_graph()
        non_directed_graph_different_edges.add_edge(labels.label_0, labels.label_0_2)
        assert non_directed_graph != non_directed_graph_different_edges
        # Checking __eq__ method for non-directed LabelGraph objects with different nodes
        non_directed_graph_different_nodes = self.non_directed_label_graph()
        non_directed_graph_different_nodes.add_node(labels.non_included_label)
        assert non_directed_graph != non_directed_graph_different_nodes
        # Checking __eq__ method by comparing non-directed LabelGraph with different type object
        assert non_directed_graph != str
        # Checks for directed LabelGraph
        # Checking __eq__ method for equal directed LabelGraph
        directed_graph = self.directed_label_graph()
        equal_directed_graph = self.directed_label_graph()
        assert directed_graph == equal_directed_graph
        # Checking __eq__ method for directed LabelGraph objects with different edges
        directed_graph_different_edges = self.directed_label_graph()
        directed_graph_different_edges.add_edge(labels.label_0_1, labels.label_0)
        assert directed_graph != directed_graph_different_edges
        # Checking __eq__ method for directed LabelGraph objects with different nodes
        directed_graph_different_nodes = self.directed_label_graph()
        directed_graph_different_nodes.add_node(labels.non_included_label)
        assert directed_graph != directed_graph_different_nodes
        # Checking __eq__ method by comparing directed LabelGraph with different type object
        assert directed_graph != str


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestLabelTree:
    @staticmethod
    def label_tree_no_children() -> LabelTree:
        label_tree = LabelTree()
        label_tree.add_edges([edges.edge_0_to_0_1, edges.edge_0_to_0_2])
        return label_tree

    @staticmethod
    def label_tree() -> LabelTree:
        label_tree = LabelTree()
        label_tree.get_labels_in_topological_order()
        # Forming Label Tree with children
        for parent, child in [
            edges.edge_0_to_0_1,
            edges.edge_0_to_0_2,
            edges.edge_0_1_to_0_1_3,
            edges.edge_0_2_to_0_2_4,
            edges.edge_0_2_to_0_2_5,
        ]:
            label_tree.add_child(parent, child)
        return label_tree

    @staticmethod
    def check_get_children_method(label_tree) -> None:
        for label, expected in [
            (labels.label_0, [labels.label_0_1, labels.label_0_2]),
            (labels.label_0_1, [labels.label_0_1_3]),
            (labels.label_0_2, [labels.label_0_2_4, labels.label_0_2_5]),
            (labels.label_0_1_3, []),
            (labels.label_0_2_4, []),
            (labels.label_0_2_5, []),
            (LabelEntity("not included", domain=Domain.CLASSIFICATION), []),
        ]:
            assert label_tree.get_children(label) == expected

    @staticmethod
    def check_get_descendants_method(label_tree) -> None:
        for label, expected in [
            (
                labels.label_0,
                [
                    labels.label_0_1,
                    labels.label_0_1_3,
                    labels.label_0_2,
                    labels.label_0_2_4,
                    labels.label_0_2_5,
                ],
            ),
            (labels.label_0_1, [labels.label_0_1_3]),
            (labels.label_0_2, [labels.label_0_2_4, labels.label_0_2_5]),
            (labels.label_0_1_3, []),
            (labels.label_0_2_4, []),
            (labels.label_0_2_5, []),
        ]:
            assert label_tree.get_descendants(label) == expected

    @staticmethod
    def check_get_ancestors_method(label_tree) -> None:
        for label, expected in [
            (labels.label_0, [labels.label_0]),
            (labels.label_0_1, [labels.label_0_1, labels.label_0]),
            (labels.label_0_2, [labels.label_0_2, labels.label_0]),
            (
                labels.label_0_1_3,
                [labels.label_0_1_3, labels.label_0_1, labels.label_0],
            ),
            (
                labels.label_0_2_4,
                [labels.label_0_2_4, labels.label_0_2, labels.label_0],
            ),
            (
                labels.label_0_2_5,
                [labels.label_0_2_5, labels.label_0_2, labels.label_0],
            ),
        ]:
            assert label_tree.get_ancestors(label) == expected

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_initialization(self):
        """
        <b>Description:</b>
        Check LabelTree class object initialization

        <b>Input data:</b>
        LabelTree object, edges and nodes to add

        <b>Expected results:</b>
        Test passes if LabelTree object "directed", "edges" and "nodes" attributes and "num_labels" and "type"
        properties and value returned by "num_nodes" method are equal expected
        """
        label_tree = LabelTree()
        # Check for initiated non-directed LabelGraph
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": label_tree.directed, "expected_value": True},
                {"attribute": label_tree.num_labels, "expected_value": 0},
                {"attribute": label_tree.num_nodes(), "expected_value": 0},
                {"attribute": label_tree.type, "expected_value": "tree"},
            ]
        )
        # Check for LabelTree with added edges and nodes
        label_tree.add_edges([edges.edge_0_to_0_1, edges.edge_0_to_0_2])
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": label_tree.directed, "expected_value": True},
                {"attribute": label_tree.num_labels, "expected_value": 3},
                {"attribute": label_tree.num_nodes(), "expected_value": 3},
                {"attribute": label_tree.type, "expected_value": "tree"},
            ]
        )
        expected_nodes = [labels.label_0, labels.label_0_1, labels.label_0_2]
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": label_tree.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        (labels.label_0, labels.label_0_1, 0, {}),
                        (labels.label_0, labels.label_0_2, 0, {}),
                    ],
                },
                {
                    "attribute": label_tree.nodes,
                    "expected_type": NodeView,
                    "expected_value": expected_nodes,
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_add_edge(self):
        """
        <b>Description:</b>
        Check LabelTree class add_edge and add_edges methods

        <b>Input data:</b>
        LabelTree object with specified directed parameters and added edges

        <b>Expected results:</b>
        Test passes if "edges" attribute of LabelTree is equal expected value after using add_edge and add_edges methods
        """
        label_tree = self.label_tree_no_children()
        # Adding edges, one of which already in LabelTree
        label_tree.topological_sort()
        label_tree.add_edges([edges.edge_0_to_0_1, edges.edge_0_1_to_0_1_3])
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": label_tree.num_labels, "expected_value": 4},
                {"attribute": label_tree.num_nodes(), "expected_value": 4},
            ]
        )
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": label_tree.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        (labels.label_0, labels.label_0_1, 0, {}),
                        (labels.label_0, labels.label_0_1, 1, {}),
                        (labels.label_0, labels.label_0_2, 0, {}),
                        (labels.label_0_1, labels.label_0_1_3, 0, {}),
                    ],
                },
                {
                    "attribute": label_tree.nodes,
                    "expected_type": NodeView,
                    "expected_value": [
                        labels.label_0,
                        labels.label_0_1,
                        labels.label_0_2,
                        labels.label_0_1_3,
                    ],
                },
            ]
        )
        assert not label_tree._LabelTree__topological_order_cache
        # Adding one existing and one non-existing edge
        label_tree.topological_sort()
        label_tree.add_edges([edges.edge_0_to_0_2, edges.edge_0_2_to_0_2_4])
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": label_tree.num_labels, "expected_value": 5},
                {"attribute": label_tree.num_nodes(), "expected_value": 5},
            ]
        )
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": label_tree.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        (labels.label_0, labels.label_0_1, 0, {}),
                        (labels.label_0, labels.label_0_1, 1, {}),
                        (labels.label_0, labels.label_0_2, 0, {}),
                        (labels.label_0, labels.label_0_2, 1, {}),
                        (labels.label_0_1, labels.label_0_1_3, 0, {}),
                        (labels.label_0_2, labels.label_0_2_4, 0, {}),
                    ],
                },
                {
                    "attribute": label_tree.nodes,
                    "expected_type": NodeView,
                    "expected_value": [
                        labels.label_0,
                        labels.label_0_1,
                        labels.label_0_2,
                        labels.label_0_1_3,
                        labels.label_0_2_4,
                    ],
                },
            ]
        )
        assert not label_tree._LabelTree__topological_order_cache

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_add_node(self):
        """
        <b>Description:</b>
        Check LabelTree class add_node method

        <b>Input data:</b>
        LabelTree object with specified directed parameter and added edges

        <b>Expected results:</b>
        Test passes if "nodes" attribute of LabelTree is equal expected value after using add_node method
        """
        label_tree = self.label_tree_no_children()
        # Adding new node
        label_tree.get_labels_in_topological_order()
        label_tree.add_node(labels.label_0_1_3)
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": label_tree.num_labels, "expected_value": 4},
                {"attribute": label_tree.num_nodes(), "expected_value": 4},
            ]
        )
        expected_edges = [
            (labels.label_0, labels.label_0_1, 0, {}),
            (labels.label_0, labels.label_0_2, 0, {}),
        ]
        expected_nodes = [
            labels.label_0,
            labels.label_0_1,
            labels.label_0_2,
            labels.label_0_1_3,
        ]
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": label_tree.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": expected_edges,
                },
                {
                    "attribute": label_tree.nodes,
                    "expected_type": NodeView,
                    "expected_value": expected_nodes,
                },
            ]
        )
        assert not label_tree._LabelTree__topological_order_cache
        # Adding existing node, only topological_order_cache should be empty
        label_tree.get_labels_in_topological_order()
        label_tree.add_node(labels.label_0)
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": label_tree.num_labels, "expected_value": 4},
                {"attribute": label_tree.num_nodes(), "expected_value": 4},
            ]
        )
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": label_tree.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": expected_edges,
                },
                {
                    "attribute": label_tree.nodes,
                    "expected_type": NodeView,
                    "expected_value": expected_nodes,
                },
            ]
        )
        assert not label_tree._LabelTree__topological_order_cache

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_clear_topological_cache(self):
        """
        <b>Description:</b>
        Check LabelTree class clear_topological_cache method

        <b>Input data:</b>
        LabelTree object with specified directed parameter and added edges

        <b>Expected results:</b>
        Test passes if "__topological_order_cache" attribute of LabelTree is equal "None" after clear_topological_cache
        """
        # Check for empty LabelTree
        label_tree = LabelTree()
        label_tree.get_labels_in_topological_order()
        label_tree.clear_topological_cache()
        assert not label_tree._LabelTree__topological_order_cache
        # Check for LabelTree with specified nodes and edges
        label_tree = self.label_tree_no_children()
        label_tree.get_labels_in_topological_order()
        label_tree.clear_topological_cache()
        assert not label_tree._LabelTree__topological_order_cache

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_relations(self):
        """
        <b>Description:</b>
        Check LabelTree class relations methods

        <b>Input data:</b>
        LabelTree object with specified directed parameter, added edges and children

        <b>Expected results:</b>
        Test passes if "get_parent", "get_children", "get_descendants", "get_siblings" and "get_ancestors" methods
        of LabelTree return expected values

        <b>Steps</b>
        1. Check add_children method
        2. Check "get_parent" method
        3. Check "get_descendants" method
        4. Check "get_siblings" method
        5. Check "get_ancestors" method
        """
        label_tree = self.label_tree()
        assert not label_tree._LabelTree__topological_order_cache
        # Checking new nodes and edges added after add_children method
        CommonGraphMethods().check_graph_non_list_attributes(
            [
                {"attribute": label_tree.num_labels, "expected_value": 6},
                {"attribute": label_tree.num_nodes(), "expected_value": 6},
            ]
        )
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": label_tree.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        (labels.label_0_1, labels.label_0, 0, {"value": None}),
                        (labels.label_0_2, labels.label_0, 0, {"value": None}),
                        (labels.label_0_1_3, labels.label_0_1, 0, {"value": None}),
                        (labels.label_0_2_4, labels.label_0_2, 0, {"value": None}),
                        (labels.label_0_2_5, labels.label_0_2, 0, {"value": None}),
                    ],
                },
                {
                    "attribute": label_tree.nodes,
                    "expected_type": NodeView,
                    "expected_value": [
                        labels.label_0_1,
                        labels.label_0,
                        labels.label_0_2,
                        labels.label_0_1_3,
                        labels.label_0_2_4,
                        labels.label_0_2_5,
                    ],
                },
            ]
        )
        # Checking "get_parent" method
        for label, expected in [
            (labels.label_0, None),
            (labels.label_0_1, labels.label_0),
            (labels.label_0_2, labels.label_0),
            (labels.label_0_1_3, labels.label_0_1),
            (labels.label_0_2_4, labels.label_0_2),
            (labels.label_0_2_5, labels.label_0_2),
        ]:
            assert label_tree.get_parent(label) == expected
        # Checking "get_children" method
        self.check_get_children_method(label_tree)
        # Checking "get_descendants" method
        self.check_get_descendants_method(label_tree)
        # Checking "get_siblings" method
        for label, expected in [
            (labels.label_0, []),
            (labels.label_0_1, [labels.label_0_2]),
            (labels.label_0_2, [labels.label_0_1]),
            (labels.label_0_1_3, []),
            (labels.label_0_2_4, [labels.label_0_2_5]),
            (labels.label_0_2_5, [labels.label_0_2_4]),
        ]:
            assert label_tree.get_siblings(label) == expected
        # Checking "get_ancestors" method
        self.check_get_ancestors_method(label_tree)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_get_labels_in_topological_order(self):
        """
        <b>Description:</b>
        Check LabelTree class get_labels_in_topological_order method

        <b>Input data:</b>
        LabelTree object with specified directed parameter, added edges and children

        <b>Expected results:</b>
        Test passes if get_labels_in_topological_order method of LabelTree returns expected value

        <b>Steps</b>
        1. Check value returned by get_labels_in_topological_order method for Tree with multiple children branches
        2. Remove node with children from tree and check value returned by get_labels_in_topological_order method
        """
        label_tree = self.label_tree()
        # Checking value returned by get_labels_in_topological_order method for tree with multiple branches
        labels_topological_order = label_tree.get_labels_in_topological_order()

        def previous_vertexes(vert_id):
            vertexes = vert_id.split("_")
            return vertexes[:-1], vertexes[-1]

        returned_vertexes = set()
        for label in labels_topological_order:
            previous, curent = previous_vertexes(label.id_)
            for vertex in previous:
                assert vertex in returned_vertexes
            returned_vertexes.add(curent)
        assert {"0", "1", "2", "3", "4", "5"} == returned_vertexes

        assert label_tree._LabelTree__topological_order_cache == labels_topological_order
        # Removing node with children and checking value returned by get_labels_in_topological_order method
        label_tree.remove_node(labels.label_0_1)
        labels_topological_order = label_tree.get_labels_in_topological_order()
        returned_vertexes = set()
        for label in labels_topological_order:
            previous, curent = previous_vertexes(label.id_)
            for vertex in previous:
                if curent != "3":  # the '1' has been removed, so that '3' is separated from the rest graph.
                    assert vertex in returned_vertexes
            returned_vertexes.add(curent)
        assert {"0", "2", "3", "4", "5"} == returned_vertexes
        assert label_tree._LabelTree__topological_order_cache == labels_topological_order

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_remove_node(self):
        """
        <b>Description:</b>
        Check LabelTree class remove_node method

        <b>Input data:</b>
        LabelTree object with specified directed parameter, added edges and children

        <b>Expected results:</b>
        Test passes if after using remove_node method on LabelTree object "edges", "nodes" and "num_labels" properties
        and "num_nodes" method return expected values

        <b>Steps</b>
        1. Check values returned by "edges", "nodes" and "num_labels" properties and "num_nodes" method after removing
        children node
        2. Check values returned by "edges", "nodes" and "num_labels" properties and "num_nodes" method after removing
        parent node
        """
        label_tree = self.label_tree()
        # Removing children node and checking "edges", "nodes" and "num_labels" properties and "num_nodes" method values
        label_tree.remove_node(labels.label_0_1_3)
        assert label_tree.num_nodes() == 5
        assert label_tree.num_labels == 5
        assert not label_tree._LabelTree__topological_order_cache
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": label_tree.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        (labels.label_0_1, labels.label_0, 0, {"value": None}),
                        (labels.label_0_2, labels.label_0, 0, {"value": None}),
                        (labels.label_0_2_4, labels.label_0_2, 0, {"value": None}),
                        (labels.label_0_2_5, labels.label_0_2, 0, {"value": None}),
                    ],
                },
                {
                    "attribute": label_tree.nodes,
                    "expected_type": NodeView,
                    "expected_value": [
                        labels.label_0_1,
                        labels.label_0,
                        labels.label_0_2,
                        labels.label_0_2_4,
                        labels.label_0_2_5,
                    ],
                },
            ]
        )
        # Removing node with children and checking value returned by get_labels_in_topological_order method
        label_tree.get_labels_in_topological_order()
        label_tree.remove_node(labels.label_0_2)
        assert label_tree.num_nodes() == 4
        assert label_tree.num_labels == 4
        assert not label_tree._LabelTree__topological_order_cache
        expected_edges = [(labels.label_0_1, labels.label_0, 0, {"value": None})]
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": label_tree.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": expected_edges,
                },
                {
                    "attribute": label_tree.nodes,
                    "expected_type": NodeView,
                    "expected_value": [
                        labels.label_0_1,
                        labels.label_0,
                        labels.label_0_2_4,
                        labels.label_0_2_5,
                    ],
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_subgraph(self):
        """
        <b>Description:</b>
        Check LabelTree class subgraph method

        <b>Input data:</b>
        LabelTree object with specified directed parameter, added edges and children

        <b>Expected results:</b>
        Test passes if LabelTree object returned by subgraph method is equal expected
        """
        label_tree = self.label_tree()
        non_included_label = LabelEntity("not included", domain=Domain.CLASSIFICATION)
        subgraph = label_tree.subgraph(
            [
                labels.label_0,
                labels.label_0_1,
                labels.label_0_2,
                labels.label_0_2_5,
                non_included_label,
            ]
        )
        assert subgraph.num_nodes() == 4
        assert subgraph.num_labels == 4
        CommonGraphMethods().check_graph_list_attributes(
            [
                {
                    "attribute": subgraph.edges,
                    "expected_type": OutMultiEdgeDataView,
                    "expected_value": [
                        (labels.label_0_1, labels.label_0, 0, {"value": None}),
                        (labels.label_0_2, labels.label_0, 0, {"value": None}),
                        (labels.label_0_2_5, labels.label_0_2, 0, {"value": None}),
                    ],
                },
                {
                    "attribute": subgraph.nodes,
                    "expected_type": NodeView,
                    "expected_value": [
                        labels.label_0_1,
                        labels.label_0,
                        labels.label_0_2,
                        labels.label_0_2_5,
                    ],
                },
            ]
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_tree_eq(self):
        """
        <b>Description:</b>
        Check LabelTree class __eq__ method

        <b>Input data:</b>
        LabelTree objects with specified directed parameter, added edges and children

        <b>Expected results:</b>
        Test passes if value returned by __eq__ method is equal expected

        <b>Steps</b>
        1. Check value returned by __eq__ method for equal LabelTree objects
        2. Check value returned by __eq__ method for LabelTree objects with different edges
        3. Check value returned by __eq__ method for LabelTree objects with different nodes
        4. Check value returned by __eq__ method for comparing LabelTree objects with different type object
        """
        label_tree = self.label_tree()
        # Checking __eq__ method for equal LabelTree objects
        equal_label_tree = self.label_tree()
        assert label_tree == equal_label_tree
        # Checking __eq__ method for LabelTree objects with different edges
        different_edges_tree = self.label_tree()
        different_edges_tree.add_edge(labels.label_0, labels.label_0_1)
        assert not label_tree == different_edges_tree
        # Checking __eq__ method for LabelTree objects with different nodes
        different_nodes_tree = self.label_tree()
        different_nodes_tree.add_node(LabelEntity("not included", domain=Domain.CLASSIFICATION))
        assert label_tree != different_nodes_tree
        # Checking __eq__ method for comparing LabelTree object with different type object
        assert label_tree != str


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestLabelSchemaEntity:
    @staticmethod
    def label_groups() -> list:
        group_1 = LabelGroup(
            name="Exclusive group 1",
            labels=[labels.label_0_1, labels.label_0_2],
            id=ID("Exclusive group 1"),
        )
        group_2 = LabelGroup(
            name="Exclusive group 2",
            labels=[labels.label_0_2_4, labels.label_0_2_5],
            id=ID("Exclusive group 2"),
        )
        return [group_1, group_2]

    @staticmethod
    def empty_labels() -> list:
        empty_label = LabelEntity(
            name="Empty label",
            domain=Domain.SEGMENTATION,
            color=Color(255, 255, 255),
            is_empty=True,
            id=ID("empty_label_1"),
        )
        empty_non_exclusive_label = LabelEntity(
            name="Empty non-exclusive label",
            domain=Domain.DETECTION,
            color=Color(255, 255, 255),
            is_empty=True,
            id=ID("empty_non_excl_label_1"),
        )
        return [empty_label, empty_non_exclusive_label]

    def empty_labels_groups(self) -> list:
        empty_labels = self.empty_labels()
        empty_group_1 = LabelGroup(
            name="Exclusive group with empty label",
            labels=[empty_labels[0]],
            id=ID("Exclusive group with empty label"),
        )
        empty_group_2 = LabelGroup(
            name="Empty label group with empty label",
            labels=[empty_labels[1]],
            group_type=LabelGroupType.EMPTY_LABEL,
            id=ID("Empty label group with empty label"),
        )
        return [empty_group_1, empty_group_2]

    def label_schema_entity(self) -> LabelSchemaEntity:
        return LabelSchemaEntity(
            label_tree=TestLabelTree.label_tree(),
            label_groups=self.label_groups() + self.empty_labels_groups(),
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_entity_default_parameters_initialization(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity object initialization

        <b>Input data:</b>
        LabelSchemaEntity object with not specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if attributes for LabelSchemaEntity with not specified "label_tree" and
        "label_groups" parameters are equal expected
        """
        empty_label_schema_entity = LabelSchemaEntity()
        assert empty_label_schema_entity.label_tree == LabelTree()
        assert empty_label_schema_entity._groups == []

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_entity_get_labels(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class get_labels method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if list returned by get_labels method is equal expected

        <b>Steps</b>
        1. Check list returned by get_labels method with include_empty parameter set to True
        2. Check list returned by get_labels method with include_empty parameter set to False
        """
        empty_labels = self.empty_labels()
        label_schema_entity = self.label_schema_entity()
        # Checking list returned by get_labels method with include_empty parameter set to True
        assert label_schema_entity.get_labels(include_empty=True) == [
            labels.label_0_1,
            labels.label_0_2,
            labels.label_0_2_4,
            labels.label_0_2_5,
            empty_labels[0],
            empty_labels[1],
        ]
        # Checking list returned by get_labels method with include_empty parameter set to False
        assert label_schema_entity.get_labels(include_empty=False) == [
            labels.label_0_1,
            labels.label_0_2,
            labels.label_0_2_4,
            labels.label_0_2_5,
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_entity_get_groups(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class get_groups method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if list returned by get_groups method is equal expected

        <b>Steps</b>
        1. Check list returned by get_groups method with include_empty parameter set to True
        2. Check list returned by get_groups method with include_empty parameter set to False
        """
        label_groups = self.label_groups()
        empty_label_groups = self.empty_labels_groups()
        label_schema_entity = self.label_schema_entity()
        # Checking list returned by get_groups method with include_empty parameter set to True
        assert label_schema_entity.get_groups(include_empty=True) == [
            label_groups[0],
            label_groups[1],
            empty_label_groups[0],
            empty_label_groups[1],
        ]
        # Checking list returned by get_groups method with include_empty parameter set to False
        assert label_schema_entity.get_groups(include_empty=False) == [
            label_groups[0],
            label_groups[1],
            empty_label_groups[0],
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_entity_add_group(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class add_group method

        <b>Input data:</b>
        LabelSchemaEntity object with label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if value returned by "get_exclusive_groups" method is equal expected

        <b>Steps</b>
        1. Check value returned by "get_exclusive_groups" method after adding group with new labels
        2. Check value returned by "get_exclusive_groups" method after adding group with single label
        3. Check value returned by "get_exclusive_groups" method after adding group with already added label
        4. Check value returned by "get_exclusive_groups" method and exclusivity of labels after adding group exclusive
        to other
        5. Check value returned by "get_exclusive_groups" method after adding non-exclusive group
        6. Check LabelGroupExistsException raised when adding LabelGroup with already existing name
        """
        empty_label_groups = self.empty_labels_groups()
        exclusive_groups = self.label_groups() + [empty_label_groups[0]]
        label_schema_entity = self.label_schema_entity()
        # Scenario for adding exclusive group with new labels
        new_exclusive_label = LabelEntity(name="New label", domain=Domain.DETECTION, id=ID("new_ex_1"))
        other_new_exclusive_label = LabelEntity(name="Other new label", domain=Domain.DETECTION, id=ID("new_ex_2"))
        new_exclusive_group = LabelGroup(
            name="New exclusive labels group",
            labels=[new_exclusive_label, other_new_exclusive_label],
            id=ID("new_ex_group"),
        )
        label_schema_entity.add_group(new_exclusive_group)
        assert label_schema_entity.get_exclusive_groups() == (exclusive_groups + [new_exclusive_group])
        # Scenario for adding exclusive group with single label
        label_schema_entity = self.label_schema_entity()
        new_exclusive_group = LabelGroup(
            name="Exclusive group with one label",
            labels=[new_exclusive_label],
            id=ID("single_excl_group"),
        )
        label_schema_entity.add_group(new_exclusive_group)
        assert label_schema_entity.get_exclusive_groups() == (exclusive_groups + [new_exclusive_group])
        # Scenario for adding exclusive group with one already existing label
        label_schema_entity = self.label_schema_entity()
        new_exclusive_group = LabelGroup(
            name="Exclusive group to link with existing",
            labels=[labels.label_0_1, new_exclusive_label],
            id=ID("new_ex_group"),
        )
        label_schema_entity.add_group(new_exclusive_group)
        assert label_schema_entity.get_exclusive_groups() == (exclusive_groups + [new_exclusive_group])
        # Scenario for adding non-exclusive group
        label_schema_entity = self.label_schema_entity()
        new_exclusive_group = LabelGroup(
            name="Non exclusive label group",
            labels=[new_exclusive_label],
            group_type=LabelGroupType.EMPTY_LABEL,
            id=ID("non_exclusive_group"),
        )
        label_schema_entity.add_group(new_exclusive_group)
        assert label_schema_entity.get_exclusive_groups() == exclusive_groups
        # Raise LabelGroupExistsException when adding LabelGroup with same name
        for group_type in [LabelGroupType.EXCLUSIVE, LabelGroupType.EMPTY_LABEL]:
            with pytest.raises(LabelGroupExistsException):
                label_schema_entity.add_group(
                    LabelGroup(
                        name="Exclusive group 1",
                        labels=[],
                        group_type=group_type,
                    )
                )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_add_child(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class add_child method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if list returned by get_children method after using add_child method are equal to expected

        <b>Steps</b>
        1. Check get_children list after using add_child method
        2. Check get_children list after using add_child method for previous pair for a second time
        """
        label_schema_entity = self.label_schema_entity()
        label_schema_entity.add_child(parent=labels.label_0_2_4, child=labels.label_0_2_5)
        assert label_schema_entity.get_children(labels.label_0_2_4) == [labels.label_0_2_5]
        label_schema_entity.add_child(parent=labels.label_0_2_4, child=labels.label_0_2_5)
        assert label_schema_entity.get_children(labels.label_0_2_4) == [labels.label_0_2_5]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_entity_get_label_ids(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class get_label_ids method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if values returned by get_label_ids method is equal expected

        <b>Steps</b>
        Check value returned by get_label_ids method for LabelSchemaEntity object
        """
        expected_non_empty_labels = [ID("0_1"), ID("0_2"), ID("0_2_4"), ID("0_2_5")]
        expected_include_empty_labels = expected_non_empty_labels + [
            ID("empty_label_1"),
            ID("empty_non_excl_label_1"),
        ]
        label_schema_entity = self.label_schema_entity()
        assert label_schema_entity.get_label_ids(include_empty=True) == expected_include_empty_labels
        assert label_schema_entity.get_label_ids(include_empty=False) == expected_non_empty_labels

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_get_label_group_by_name(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class get_label_group_by_name method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if value returned by get_label_group_by_name method is equal expected

        <b>Steps</b>
        1. Check get_label_group_by_name method for searching exclusive group
        2. Check get_label_group_by_name method for searching empty label group
        2. Check get_label_group_by_name method for searching non_existing group
        """
        label_schema_entity = self.label_schema_entity()
        label_groups = self.label_groups()
        empty_non_excl_group = self.empty_labels_groups()[1]
        # Checking get_label_group_by_name method for searching exclusive group for not specified empty labels
        assert label_schema_entity.get_label_group_by_name("Exclusive group 1") == label_groups[0]
        # Checking get_label_group_by_name method for searching empty label group with empty label
        assert label_schema_entity.get_label_group_by_name("Empty label group with empty label") == empty_non_excl_group
        # Checking get_label_group_by_name method for searching non-existing group
        assert not label_schema_entity.get_label_group_by_name("Non-existing group")

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_get_exclusive_groups(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class get_exclusive_groups method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if value returned by get_exclusive_groups method is equal expected

        <b>Steps</b>
        1. Check get_exclusive_groups method
        2. Check get_exclusive_groups method after adding exclusive group
        3. Check get_exclusive_groups method after adding empty label group
        """
        label_schema_entity = self.label_schema_entity()
        exclusive_groups = self.label_groups() + [self.empty_labels_groups()[0]]
        # Checking get_exclusive_groups method for searching exclusive groups
        assert label_schema_entity.get_exclusive_groups() == exclusive_groups
        # Checking get_exclusive_groups method after adding new exclusive group
        new_label = LabelEntity(
            name="New label",
            domain=Domain.DETECTION,
            color=Color(100, 16, 25),
            id=ID("new_ex_1"),
        )
        new_labels_group = LabelGroup(
            name="New exclusive labels group",
            labels=[new_label],
            id=ID("new_ex_group"),
        )
        label_schema_entity.add_group(new_labels_group)
        exclusive_groups.append(new_labels_group)
        assert label_schema_entity.get_exclusive_groups() == exclusive_groups
        # Checking get_exclusive_groups method after adding empty label group
        empty_label_group = LabelGroup(
            name="New non-exclusive labels group",
            labels=[new_label],
            group_type=LabelGroupType.EMPTY_LABEL,
            id=ID("new_ex_group"),
        )
        label_schema_entity.add_group(empty_label_group)
        assert label_schema_entity.get_exclusive_groups() == exclusive_groups

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_add_labels_to_group_by_group_name(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class add_labels_to_group_by_group_name method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if labels attribute returned by group which specified in add_labels_to_group_by_group_name method
        is equal expected

        <b>Steps</b>
        1. Check add_labels_to_group_by_group_name method to add labels to exclusive group
        2. Check add_labels_to_group_by_group_name method to add labels to empty label group
        3. Check LabelGroupDoesNotExistException raised when adding labels to non-existing group
        """
        label_schema_entity = self.label_schema_entity()
        non_exclusive_label = self.empty_labels()[1]
        # Checking add_labels_to_group_by_group_name method to add labels to exclusive group
        new_label = LabelEntity(
            name="New label",
            domain=Domain.DETECTION,
            color=Color(100, 16, 25),
            id=ID("new_ex_1"),
        )
        new_empty_label = LabelEntity(
            name="New empty label",
            domain=Domain.DETECTION,
            color=Color(81, 100, 10),
            id=ID("new_ex_2"),
        )
        exclusive_group_name = "Exclusive group 1"
        label_schema_entity.add_labels_to_group_by_group_name(
            group_name=exclusive_group_name, labels=[new_label, new_empty_label]
        )
        assert label_schema_entity.get_label_group_by_name(exclusive_group_name).labels == [
            labels.label_0_1,
            labels.label_0_2,
            new_label,
            new_empty_label,
        ]
        # Checking add_labels_to_group_by_group_name method to add labels to non-exclusive group
        new_empty_label = LabelEntity(
            name="New non-exclusive empty_label",
            domain=Domain.SEGMENTATION,
            is_empty=True,
            id=ID("empty_label_1"),
        )
        empty_label_group_name = "Empty label group with empty label"
        label_schema_entity.add_labels_to_group_by_group_name(
            group_name=empty_label_group_name, labels=[new_empty_label]
        )
        assert label_schema_entity.get_label_group_by_name(empty_label_group_name).labels == [
            non_exclusive_label,
            new_empty_label,
        ]
        # Checking that LabelGroupDoesNotExistException raised when adding labels to non-existing group
        with pytest.raises(LabelGroupDoesNotExistException):
            label_schema_entity.add_labels_to_group_by_group_name(
                group_name="Non-existing group", labels=[new_empty_label]
            )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_relations(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity relations methods

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if "get_children", "get_descendants" and "get_ancestors" methods
        of LabelTree return expected values

        <b>Steps</b>
        1. Check "get_children" method
        2. Check "get_descendants" method
        3. Check "get_ancestors" method
        """
        label_schema_entity = self.label_schema_entity()
        # Checking get_children method
        TestLabelTree.check_get_children_method(label_schema_entity)
        # Checking get_descendants method
        TestLabelTree.check_get_descendants_method(label_schema_entity)
        # Checking get_ancestors method
        TestLabelTree.check_get_ancestors_method(label_schema_entity)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_get_group_containing_label(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class get_group_containing_label method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if value returned by get_group_containing_label method is equal expected

        <b>Steps</b>
        1. Check get_group_containing_label method for label included in exclusive group
        2. Check get_group_containing_label method for label included in empty label
        3. Check get_group_containing_label method for label not included in any group
        """
        label_schema_entity = self.label_schema_entity()
        label_groups = self.label_groups()
        # Checking get_group_containing_label method for label included in exclusive group
        assert label_schema_entity.get_group_containing_label(labels.label_0_1) == label_groups[0]
        assert label_schema_entity.get_group_containing_label(labels.label_0_2) == label_groups[0]
        assert label_schema_entity.get_group_containing_label(labels.label_0_2_4) == label_groups[1]
        assert label_schema_entity.get_group_containing_label(labels.label_0_2_5) == label_groups[1]
        # Checking get_group_containing_label method for label included in non-exclusive group
        assert label_schema_entity.get_group_containing_label(self.empty_labels()[1]) == self.empty_labels_groups()[1]
        # Checking get_group_containing_label method for label not included in any group
        assert not label_schema_entity.get_group_containing_label(labels.label_0)
        assert not label_schema_entity.get_group_containing_label(labels.label_0_1_3)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_get_label(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class __get_label method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if LabelEntity object returned by __get_label method is equal expected

        <b>Steps</b>
        1. Check __get_label method for searching ScoredLabel object
        2. Check __get_label method for searching LabelEntity object
        3. Check __get_label method for searching non-label object
        """
        label_schema_entity = self.label_schema_entity()
        # Checking __get_label for searching ScoredLabel object
        label_to_set_scored = LabelEntity(
            name="Scored label",
            domain=Domain.DETECTION,
            color=Color(red=100, green=50, blue=200),
            id=ID("scored_label_1"),
        )
        scored_label = ScoredLabel(label=label_to_set_scored)
        scored_labels_group = LabelGroup(name="Group with scored label", labels=[scored_label])
        label_schema_entity.add_group(scored_labels_group)
        assert label_schema_entity._LabelSchemaEntity__get_label(scored_label) == scored_label.get_label()
        # Checking __get_label for searching LabelEntity object
        assert label_schema_entity._LabelSchemaEntity__get_label(labels.label_0) == labels.label_0
        # Checking __get_label method for searching non-label object
        with pytest.raises(ValueError):
            label_schema_entity._LabelSchemaEntity__get_label(str)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_repr(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class __repr__ method

        <b>Input data:</b>
        LabelSchemaEntity object with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if value returned by __repr__ method is equal expected
        """
        label_schema_entity = self.label_schema_entity()
        label_groups = self.label_groups() + self.empty_labels_groups()
        assert repr(label_schema_entity) == f"LabelSchemaEntity(label_groups={label_groups})"

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_eq(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class __eq__ method

        <b>Input data:</b>
        LabelSchemaEntity objects with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if value returned by __eq__ method is equal expected

        <b>Steps</b>
        1. Check __eq__ method for equal LabelSchemaEntity objects
        2. Check __eq__ method for LabelSchemaEntity objects with unequal label_tree
        3. Check __eq__ method for LabelSchemaEntity objects with unequal LabelGroups
        4. Check __eq__ method for comparing LabelSchemaEntity object object of other type
        """
        # Checking __eq__ method for equal LabelSchemaEntity objects
        label_schema_entity = self.label_schema_entity()
        equal_label_schema_entity = self.label_schema_entity()
        assert label_schema_entity == equal_label_schema_entity
        # Checking __eq__ method for equal LabelSchemaEntity objects with unequal label_tree
        unequal_tree_label_schema_entity = self.label_schema_entity()
        unequal_tree_label_schema_entity.label_tree.add_edge(labels.label_0_1_3, labels.label_0_2_4)
        assert label_schema_entity != unequal_tree_label_schema_entity
        # Checking __eq__ method for equal LabelSchemaEntity objects with unequal LabelGroups
        unequal_groups_label_schema_entity = self.label_schema_entity()
        unequal_groups_label_schema_entity.add_labels_to_group_by_group_name("Exclusive group 1", [labels.label_0])
        assert label_schema_entity != unequal_groups_label_schema_entity
        # Checking __eq__ method for comparing LabelSchemaEntity object object of other type
        assert label_schema_entity != str

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_from_labels(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class from_labels method

        <b>Input data:</b>
        LabelSchemaEntity objects with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if LabelSchemaEntity object returned by from_labels method is equal expected
        """
        expected_labels = [
            labels.label_0,
            labels.label_0_1,
            labels.label_0_2,
            labels.label_0_2_4,
        ]
        label_schema_entity = LabelSchemaEntity.from_labels(expected_labels)
        labels_schema_entity_groups = label_schema_entity._groups
        assert isinstance(label_schema_entity, LabelSchemaEntity)
        assert label_schema_entity.label_tree == LabelTree()
        assert len(labels_schema_entity_groups) == 1
        assert labels_schema_entity_groups[0].name == "from_label_list"
        assert labels_schema_entity_groups[0].labels == expected_labels

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_label_schema_resolve_labels_probabilistic(self):
        """
        <b>Description:</b>
        Check LabelSchemaEntity class resolve_labels_probabilistic method

        <b>Input data:</b>
        LabelSchemaEntity objects with specified label_tree and label_groups parameters

        <b>Expected results:</b>
        Test passes if LabelSchemaEntity object returned by resolve_labels_probabilistic
        method is equal expected
        """
        label_schema = LabelSchemaEntity()
        labels_1 = [
            LabelEntity(
                name=f"Label {i}",
                domain=Domain.CLASSIFICATION,
                id=ID(f"{i}"),
                color=Color(i, 200, 166),
            )
            for i in range(2)
        ]
        labels_2 = [
            LabelEntity(
                name=f"Label {i}",
                domain=Domain.CLASSIFICATION,
                id=ID(f"{i}"),
                color=Color(i, 200, 166),
            )
            for i in range(2, 4)
        ]
        labels_3 = [
            LabelEntity(
                name=f"Label {i}",
                domain=Domain.CLASSIFICATION,
                id=ID(f"{i}"),
                color=Color(i, 200, 166),
            )
            for i in range(4, 5)
        ]
        group_1 = LabelGroup(name="labels1", labels=labels_1, group_type=LabelGroupType.EXCLUSIVE)
        label_schema.add_group(group_1)
        group_2 = LabelGroup(name="labels2", labels=labels_2, group_type=LabelGroupType.EXCLUSIVE)
        label_schema.add_group(group_2)

        label_schema.add_group(LabelGroup("child_labels_2", labels_3, LabelGroupType.EXCLUSIVE))
        label_schema.add_child(labels_2[0], labels_3[0])

        assert 1 == len(label_schema.get_descendants(labels_2[0]))

        # supress non-maximum labels
        predicted_labels = [
            ScoredLabel(labels_1[0], 0.1),
            ScoredLabel(labels_1[1], 0.5),
            ScoredLabel(labels_2[0], 0.2),
            ScoredLabel(labels_2[1], 0.5),
        ]
        resloved_labels = label_schema.resolve_labels_probabilistic(predicted_labels)
        assert [
            ScoredLabel(labels_1[1], 0.5),
            ScoredLabel(labels_2[1], 0.5),
        ] == resloved_labels

        # supress children of non-maximum labels
        predicted_labels = [
            ScoredLabel(labels_2[0], 0.1),
            ScoredLabel(labels_2[1], 0.5),
            ScoredLabel(labels_3[0], 0.4),
        ]
        resloved_labels = label_schema.resolve_labels_probabilistic(predicted_labels)
        assert [ScoredLabel(labels_2[1], 0.5)] == resloved_labels



# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
import tempfile
from datetime import datetime
from pathlib import Path

import pytest

from otx.api.configuration import ConfigurableParameters, cfg_helper
from otx.api.entities.annotation import NullAnnotationSceneEntity
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.metrics import NullPerformance, Performance, ScoreMetric
from otx.api.entities.model import (
    ModelConfiguration,
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.model_template import TargetDevice, parse_model_template
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.adapters.model_adapter import ModelAdapter
from otx.api.utils.time_utils import now
from tests.test_helpers import generate_random_single_image
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestModelPrecision:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_precision(self):
        """
        <b>Description:</b>
        Check that ModelPrecision correctly returns the precision name

        <b>Expected results:</b>
        Test passes if ModelPrecision correctly returns the precision name

        <b>Steps</b>
        1. Check precisions in the ModelPrecision
        """

        model_precision = ModelPrecision
        assert len(model_precision) == 4


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestModelFormat:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_format(self):
        """
        <b>Description:</b>
        Check that ModelFormat correctly returns the format name

        <b>Expected results:</b>
        Test passes if ModelFormat correctly returns the format name

        <b>Steps</b>
        1. Check formats in the ModelFormat
        """

        model_format = ModelFormat
        assert len(model_format) == 3


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestModelOptimizationType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_optimization_type(self):
        """
        <b>Description:</b>
        Check that ModelOptimizationType correctly returns the optimization type name

        <b>Expected results:</b>
        Test passes if ModelOptimizationType correctly returns the optimization type name

        <b>Steps</b>
        1. Check optimization types in the ModelOptimizationType
        """

        model_optimization_type = ModelOptimizationType
        assert len(model_optimization_type) == 4


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestOptimizationMethod:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_optimization_method(self):
        """
        <b>Description:</b>
        Check that OptimizationMethod correctly returns the optimization method name

        <b>Expected results:</b>
        Test passes if OptimizationMethod correctly returns the optimization method name

        <b>Steps</b>
        1. Check optimization methods in the OptimizationMethod
        """

        optimization_method = OptimizationMethod
        assert len(optimization_method) == 2


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestModelConfiguration:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_configuration(self):
        """
        <b>Description:</b>
        Check that ModelConfiguration correctly returns the configuration

        <b>Input data:</b>
        ConfigurableParameters, LabelSchemaEntity

        <b>Expected results:</b>
        Test passes if ModelConfiguration correctly returns the configuration

        <b>Steps</b>
        1. Check configuration params in the ModelConfiguration
        """
        parameters = ConfigurableParameters(header="Test header")
        label_schema = LabelSchemaEntity()
        model_configuration = ModelConfiguration(configurable_parameters=parameters, label_schema=label_schema)
        assert model_configuration.configurable_parameters == parameters
        assert model_configuration.get_label_schema() == label_schema


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestModelEntity:
    creation_date = now()

    def generate_random_image(self):
        with generate_random_single_image() as path:
            image = Image(file_path=path)
            return DatasetItemEntity(media=image, annotation_scene=NullAnnotationSceneEntity())

    def dataset(self):
        return DatasetEntity(items=[self.generate_random_image()])

    def configuration(self):
        parameters = ConfigurableParameters(header="Test header")
        label_schema = LabelSchemaEntity()
        return ModelConfiguration(configurable_parameters=parameters, label_schema=label_schema)

    def other_configuration(self):
        parameters = ConfigurableParameters(header="Other test header")
        label_schema = LabelSchemaEntity()
        return ModelConfiguration(configurable_parameters=parameters, label_schema=label_schema)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_entity_default_values(self):
        """
        <b>Description:</b>
        Check that ModelEntity correctly returns the default values

        <b>Expected results:</b>
        Test passes if ModelEntity correctly returns the default values

        <b>Steps</b>
        1. Check default values in the ModelEntity
        """

        model_entity = ModelEntity(train_dataset=self.dataset(), configuration=self.configuration())

        assert model_entity.id_ == ID()
        assert type(model_entity.configuration) == ModelConfiguration
        assert type(model_entity.creation_date) == datetime
        assert type(model_entity.train_dataset) == DatasetEntity
        assert model_entity.version == 1
        assert model_entity.model_format == ModelFormat.OPENVINO
        assert model_entity.precision == [ModelPrecision.FP32]
        assert model_entity.target_device == TargetDevice.CPU
        assert model_entity.optimization_type == ModelOptimizationType.NONE
        assert model_entity.performance == NullPerformance()

        for default_val_none in [
            "previous_trained_revision",
            "previous_revision",
            "target_device_type",
        ]:
            assert getattr(model_entity, default_val_none) is None

        for default_val_0_0 in ["training_duration", "model_size_reduction"]:
            assert getattr(model_entity, default_val_0_0) == 0.0

        for default_val_empty_list in ["tags", "optimization_methods"]:
            assert getattr(model_entity, default_val_empty_list) == []

        for default_val_empty_dict in [
            "model_adapters",
            "optimization_objectives",
            "performance_improvement",
        ]:
            assert getattr(model_entity, default_val_empty_dict) == {}

        for default_val_zero in ["latency", "fps_throughput"]:
            assert getattr(model_entity, default_val_zero) == 0

        assert model_entity.is_optimized() is False

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_entity_sets_values(self):
        """
        <b>Description:</b>
        Check that ModelEntity correctly returns the set values

        <b>Expected results:</b>
        Test passes if ModelEntity correctly returns the set values

        <b>Steps</b>
        1. Check set values in the ModelEntity
        """

        def __get_path_to_file(filename: str):
            """
            Return the path to the file named 'filename', which lives in the tests/entities directory
            """
            return str(Path(__file__).parent / Path(filename))

        car = LabelEntity(name="car", domain=Domain.DETECTION)
        labels_list = [car]
        dummy_template = __get_path_to_file("./dummy_template.yaml")
        model_template = parse_model_template(dummy_template)
        hyper_parameters = model_template.hyper_parameters.data
        params = cfg_helper.create(hyper_parameters)
        labels_schema = LabelSchemaEntity.from_labels(labels_list)
        environment = TaskEnvironment(
            model=None,
            hyper_parameters=params,
            label_schema=labels_schema,
            model_template=model_template,
        )

        item = self.generate_random_image()
        dataset = DatasetEntity(items=[item])
        score_metric = ScoreMetric(name="Model accuracy", value=0.5)

        model_entity = ModelEntity(train_dataset=self.dataset(), configuration=self.configuration())

        set_params = {
            "configuration": environment.get_model_configuration(),
            "train_dataset": dataset,
            "id": ID(1234567890),
            "creation_date": self.creation_date,
            "previous_trained_revision": 5,
            "previous_revision": 2,
            "version": 2,
            "tags": ["tree", "person"],
            "model_format": ModelFormat.BASE_FRAMEWORK,
            "performance": Performance(score_metric),
            "training_duration": 5.8,
            "precision": [ModelPrecision.INT8],
            "latency": 328,
            "fps_throughput": 20,
            "target_device": TargetDevice.GPU,
            "target_device_type": "notebook",
            "optimization_methods": [OptimizationMethod.QUANTIZATION],
            "optimization_type": ModelOptimizationType.MO,
            "optimization_objectives": {"param": "Test param"},
            "performance_improvement": {"speed", 0.5},
            "model_size_reduction": 1.0,
        }

        for key, value in set_params.items():
            setattr(model_entity, key, value)
            assert getattr(model_entity, key) == value

        assert model_entity.is_optimized() is True

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_entity_model_adapters(self):
        """
        <b>Description:</b>
        Check that ModelEntity correctly returns the adapters

        <b>Expected results:</b>
        Test passes if ModelEntity correctly returns the adapters

        <b>Steps</b>
        1. Create a ModelEntity with adapters
        2. Change data source for an adapter
        3. Remove an adapter
        """

        data_source_0 = b"{0: binaryrepo://localhost/repo/data_source/0}"
        data_source_1 = b"binaryrepo://localhost/repo/data_source/1"
        data_source_2 = b"binaryrepo://localhost/repo/data_source/2"
        data_source_3 = b"binaryrepo://localhost/repo/data_source/3"

        temp_dir = tempfile.TemporaryDirectory()
        temp_file = os.path.join(temp_dir.name, "data_source_0")

        with open(temp_file, "wb") as tmp:
            tmp.write(data_source_0)

        model_adapters = {
            "0": ModelAdapter(data_source=data_source_0),
            "1": ModelAdapter(data_source=data_source_1),
            "2": ModelAdapter(data_source=data_source_2),
        }

        model_entity = ModelEntity(
            train_dataset=self.dataset(),
            configuration=self.configuration(),
            model_adapters=model_adapters,
        )

        assert model_entity.weight_paths == {}

        # Adapter with key 0 not from file
        assert model_entity.model_adapters["0"].from_file_storage is False

        model_entity.set_data("0", temp_file)

        for adapter in model_entity.model_adapters:
            if adapter == "0":
                # Adapter with key 0 from file
                assert model_entity.model_adapters[adapter].from_file_storage is True
            else:
                assert model_entity.model_adapters[adapter].from_file_storage is False

        assert model_entity.get_data("1") == data_source_1

        model_entity.set_data("2", data_source_1)
        assert model_entity.get_data("2") == data_source_1
        assert len(model_entity.model_adapters) == 3

        model_entity.set_data("3", data_source_3)
        assert model_entity.get_data("3") == data_source_3
        assert len(model_entity.model_adapters) == 4

        model_entity.delete_data("3")
        assert len(model_entity.model_adapters) == 3

        # Attempt to retrieve a missing and deleted key
        with pytest.raises(KeyError):
            model_entity.get_data("5")

        with pytest.raises(KeyError):
            model_entity.get_data("3")

        # The presence of outdated code in ModelEntity "model_adapter.data_source.binary_url"
        with pytest.raises(AttributeError):
            assert model_entity.weight_paths == {"0": temp_file}

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_entity__eq__(self):
        """
        <b>Description:</b>
        Check that ModelEntity __eq__ method

        <b>Expected results:</b>
        Test passes if ModelEntity equal ModelEntity and not equal another type
        """
        dataset = self.dataset()
        other_model_entity = ModelEntity(train_dataset=dataset, configuration=self.configuration())
        model_entity = ModelEntity(train_dataset=dataset, configuration=self.configuration())
        third_model_entity = ModelEntity(train_dataset=self.dataset(), configuration=self.other_configuration())
        assert model_entity.__eq__("") is False
        assert model_entity == other_model_entity
        assert model_entity != third_model_entity



"""This UnitTest tests ShapeFactory functionality"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import pytest

from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.utils.shape_factory import ShapeFactory
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestShapeFactory:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_shape_conversion(self):
        """
        <b>Description:</b>
        Checks that conversions from Rectangle to other shapes works correctly

        <b>Input data:</b>
        A rectangle at [0.25, 0.1, 0.5, 0.3]

        <b>Expected results:</b>
        The test passes if the rectangle can be converted to Ellipse and Polygon

        <b>Steps</b>
        1. Create rectangle and get coordinates
        2. Convert to Ellipse
        3. Convert to Polygon
        4. Convert to Rectangle
        """
        rectangle = Rectangle(x1=0.25, y1=0.1, x2=0.5, y2=0.3)
        rectangle_coords = (rectangle.x1, rectangle.y1, rectangle.x2, rectangle.y2)

        ellipse = ShapeFactory.shape_as_ellipse(rectangle)
        assert isinstance(ellipse, Ellipse)
        assert (ellipse.x1, ellipse.y1, ellipse.x2, ellipse.y2) == rectangle_coords

        polygon = ShapeFactory.shape_as_polygon(rectangle)
        assert isinstance(polygon, Polygon)
        assert (
            polygon.min_x,
            polygon.min_y,
            polygon.max_x,
            polygon.max_y,
        ) == rectangle_coords

        rectangle2 = ShapeFactory.shape_as_rectangle(rectangle)
        assert isinstance(rectangle2, Rectangle)
        assert rectangle == rectangle2

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_ellipse_shape_conversion(self):
        """
        <b>Description:</b>
        Checks that conversions from Ellipse to other shapes works correctly

        <b>Input data:</b>
        A rectangle at [0.1, 0.1, 0.5, 0.2]

        <b>Expected results:</b>
        The test passes if the Ellipse can be converted to Rectangle and Polygon

        <b>Steps</b>
        1. Create Ellipse and get coordinates
        2. Convert to Ellipse
        3. Convert to Polygon
        4. Convert to Rectangle
        """
        ellipse = Ellipse(x1=0.1, y1=0.1, x2=0.5, y2=0.2)
        ellipse_coords = (ellipse.x1, ellipse.y1, ellipse.x2, ellipse.y2)

        ellipse2 = ShapeFactory.shape_as_ellipse(ellipse)
        assert isinstance(ellipse, Ellipse)
        assert ellipse == ellipse2

        polygon = ShapeFactory.shape_as_polygon(ellipse)
        assert isinstance(polygon, Polygon)
        assert polygon.min_x == pytest.approx(ellipse.x1, 0.1)
        assert polygon.min_y == pytest.approx(ellipse.y1, 0.1)
        assert polygon.max_x == pytest.approx(ellipse.x2, 0.1)
        assert polygon.max_y == pytest.approx(ellipse.y2, 0.1)

        rectangle = ShapeFactory.shape_as_rectangle(ellipse)
        assert isinstance(rectangle, Rectangle)
        assert (
            rectangle.x1,
            rectangle.y1,
            rectangle.x2,
            rectangle.y2,
        ) == ellipse_coords

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_polygon_shape_conversion(self):
        """
        <b>Description:</b>
        Checks that conversions from Polygon to other shapes works correctly

        <b>Input data:</b>
        A Polygon at [[0.01, 0.2], [0.35, 0.2], [0.35, 0.4]]

        <b>Expected results:</b>
        The test passes if the Polygon can be converted to Rectangle and Ellipse

        <b>Steps</b>
        1. Create rectangle and get coordinates
        2. Convert to Ellipse
        3. Convert to Polygon
        4. Convert to Rectangle
        """
        point1 = Point(x=0.01, y=0.2)
        point2 = Point(x=0.35, y=0.2)
        point3 = Point(x=0.35, y=0.4)
        polygon = Polygon(points=[point1, point2, point3])
        polygon_coords = (polygon.min_x, polygon.min_y, polygon.max_x, polygon.max_y)

        ellipse = ShapeFactory.shape_as_ellipse(polygon)
        assert isinstance(ellipse, Ellipse)
        assert (ellipse.x1, ellipse.y1, ellipse.x2, ellipse.y2) == polygon_coords

        polygon2 = ShapeFactory.shape_as_polygon(polygon)
        assert isinstance(polygon2, Polygon)
        assert polygon == polygon2

        rectangle = ShapeFactory.shape_as_rectangle(polygon)
        assert isinstance(rectangle, Rectangle)
        assert (
            rectangle.x1,
            rectangle.y1,
            rectangle.x2,
            rectangle.y2,
        ) == polygon_coords

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_produces_valid_crop(self):
        """
        <b>Description:</b>
        Checks that shape_produces_valid_crop returns the correct values and
        does not raise errors

        <b>Input data:</b>
        A valid Rectangle at [0, 0.4, 1, 0.5]
        A Polygon that has an invalid bounding box

        <b>Expected results:</b>
        The test passes if the call with the Rectangle returns True and
        the one with the polygon returns False

        <b>Steps</b>
        1. Check Valid Rectangle
        2. Check invalid Polygon
        """
        rectangle = Rectangle(x1=0, y1=0.4, x2=1, y2=0.5)
        assert ShapeFactory.shape_produces_valid_crop(rectangle, 100, 100)

        point1 = Point(x=0.01, y=0.1)
        point2 = Point(x=0.35, y=0.1)
        point3 = Point(x=0.35, y=0.1)
        polygon = Polygon(points=[point1, point2, point3])
        assert not ShapeFactory.shape_produces_valid_crop(polygon, 100, 250)



"""Task of OTX Detection."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, List, Optional

import cv2
import numpy as np
import pycocotools.mask as mask_util
import torch
from mmcv.utils import ConfigDict

from otx.algorithms.common.tasks.base_task import TRAIN_TYPE_DIR_PATH, OTXTask
from otx.algorithms.common.utils.callback import (
    InferenceProgressCallback,
    TrainingProgressCallback,
)
from otx.algorithms.common.utils.ir import embed_ir_model_data
from otx.algorithms.common.utils.logger import get_logger
from otx.algorithms.detection.configs.base import DetectionConfig
from otx.algorithms.detection.utils import get_det_model_api_configuration
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import config_to_bytes, ids_to_strings
from otx.api.entities.annotation import Annotation
from otx.api.entities.datasets import DatasetEntity, DatasetPurpose
from otx.api.entities.explain_parameters import ExplainParameters
from otx.api.entities.id import ID
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    MetricsGroup,
    ScoreMetric,
    VisualizationType,
)
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
)
from otx.api.entities.model_template import TaskType
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.entities.train_parameters import TrainParameters, default_progress_callback
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.utils.dataset_utils import add_saliency_maps_to_dataset_item
from otx.cli.utils.multi_gpu import is_multigpu_child_process

logger = get_logger()


class OTXDetectionTask(OTXTask, ABC):
    """Task class for OTX detection."""

    # pylint: disable=too-many-instance-attributes, too-many-locals
    def __init__(self, task_environment: TaskEnvironment, output_path: Optional[str] = None):
        super().__init__(task_environment, output_path)
        self._task_config = DetectionConfig
        self._hyperparams: ConfigDict = task_environment.get_hyper_parameters(self._task_config)
        self._train_type = self._hyperparams.algo_backend.train_type
        self._model_dir = os.path.join(
            os.path.abspath(os.path.dirname(self._task_environment.model_template.model_template_path)),
            TRAIN_TYPE_DIR_PATH[self._train_type.name],
        )
        self._anchors: Dict[str, int] = {}

        if hasattr(self._hyperparams, "postprocessing") and hasattr(
            self._hyperparams.postprocessing, "confidence_threshold"
        ):
            self.confidence_threshold = self._hyperparams.postprocessing.confidence_threshold
        else:
            self.confidence_threshold = 0.0

        if task_environment.model is not None:
            self._load_model()

        if self._hyperparams.tiling_parameters.enable_tiling:
            self.data_pipeline_path = os.path.join(self._model_dir, "tile_pipeline.py")
        else:
            self.data_pipeline_path = os.path.join(self._model_dir, "data_pipeline.py")

    def _load_tiling_parameters(self, model_data):
        """Load tiling parameters from PyTorch model.

        Args:
            model_data: The model data.

        Raises:
            RuntimeError: If tile classifier is enabled but not found in the trained model.
        """
        loaded_tiling_parameters = model_data.get("config", {}).get("tiling_parameters", {})
        if loaded_tiling_parameters.get("enable_tiling", {}).get("value", False):
            logger.info("Load tiling parameters")
            hparams = self._hyperparams.tiling_parameters
            hparams.enable_tiling = loaded_tiling_parameters["enable_tiling"]["value"]
            hparams.tile_size = loaded_tiling_parameters["tile_size"]["value"]
            hparams.tile_overlap = loaded_tiling_parameters["tile_overlap"]["value"]
            hparams.tile_max_number = loaded_tiling_parameters["tile_max_number"]["value"]
            # check backward compatibility
            enable_tile_classifier = loaded_tiling_parameters.get("enable_tile_classifier", {}).get("value", False)
            if enable_tile_classifier:
                found_tile_classifier = any(
                    layer_name.startswith("tile_classifier") for layer_name in model_data["model"]["state_dict"].keys()
                )
                if not found_tile_classifier:
                    raise RuntimeError(
                        "Tile classifier is enabled but not found in the trained model. Please retrain your model."
                    )
                hparams.enable_tile_classifier = loaded_tiling_parameters["enable_tile_classifier"]["value"]

    def _load_model_ckpt(self, model: Optional[ModelEntity]) -> Optional[Dict]:
        """Load model checkpoint from model entity.

        Args:
            model (Optional[ModelEntity]): The model entity.

        Returns:
            dict: The model checkpoint including model weights and other parameters.
        """
        if model and "weights.pth" in model.model_adapters:
            # If a model has been trained and saved for the task already, create empty model and load weights here
            buffer = io.BytesIO(model.get_data("weights.pth"))
            model_data = torch.load(buffer, map_location=torch.device("cpu"))

            # set confidence_threshold as well
            self.confidence_threshold = model_data.get("confidence_threshold", self.confidence_threshold)
            if model_data.get("anchors"):
                self._anchors = model_data["anchors"]
            self._load_tiling_parameters(model_data)
            return model_data
        return None

    def train(
        self, dataset: DatasetEntity, output_model: ModelEntity, train_parameters: Optional[TrainParameters] = None
    ):
        """Train function for OTX detection task.

        Actual training is processed by _train_model fucntion
        """
        logger.info("train()")
        # Check for stop signal when training has stopped.
        # If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # Set OTX LoggerHook & Time Monitor
        if train_parameters:
            update_progress_callback = train_parameters.update_progress
        else:
            update_progress_callback = default_progress_callback
        self._time_monitor = TrainingProgressCallback(update_progress_callback)

        dataset.purpose = DatasetPurpose.TRAINING
        results = self._train_model(dataset)

        # Check for stop signal when training has stopped. If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # get output model
        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            # output_model.model_status = ModelStatus.FAILED
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt

        # get prediction on validation set
        self._is_training = False
        val_dataset = dataset.get_subset(Subset.VALIDATION)
        val_preds, val_map = self._infer_model(val_dataset, InferenceParameters(is_evaluation=True))

        preds_val_dataset = val_dataset.with_empty_annotations()
        self._add_predictions_to_dataset(val_preds, preds_val_dataset, 0.0)

        result_set = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=val_dataset,
            prediction_dataset=preds_val_dataset,
        )

        # adjust confidence threshold
        if self._hyperparams.postprocessing.result_based_confidence_threshold:
            best_confidence_threshold = None
            logger.info("Adjusting the confidence threshold")
            metric = MetricsHelper.compute_f_measure(result_set, vary_confidence_threshold=True)
            if metric.best_confidence_threshold:
                best_confidence_threshold = metric.best_confidence_threshold.value
            if best_confidence_threshold is None:
                raise ValueError("Cannot compute metrics: Invalid confidence threshold!")
            logger.info(f"Setting confidence threshold to {best_confidence_threshold} based on results")
            self.confidence_threshold = best_confidence_threshold
        else:
            metric = MetricsHelper.compute_f_measure(result_set, vary_confidence_threshold=False)

        # compose performance statistics
        # TODO[EUGENE]: HOW TO ADD A MAE CURVE FOR TaskType.COUNTING?
        performance = metric.get_performance()
        performance.dashboard_metrics.extend(self._generate_training_metrics(self._learning_curves, val_map))
        logger.info(f"Final model performance: {str(performance)}")
        # save resulting model
        self.save_model(output_model)
        output_model.performance = performance
        logger.info("train done.")

    @abstractmethod
    def _train_model(self, dataset: DatasetEntity):
        """Train model and return the results."""
        raise NotImplementedError

    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main infer function."""
        logger.info("infer()")
        process_saliency_maps = False
        explain_predicted_classes = True

        update_progress_callback = default_progress_callback
        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress  # type: ignore
            process_saliency_maps = inference_parameters.process_saliency_maps
            explain_predicted_classes = inference_parameters.explain_predicted_classes

        self._time_monitor = InferenceProgressCallback(len(dataset), update_progress_callback)
        # If confidence threshold is adaptive then up-to-date value should be stored in the model
        # and should not be changed during inference. Otherwise user-specified value should be taken.
        if not self._hyperparams.postprocessing.result_based_confidence_threshold:
            self.confidence_threshold = self._hyperparams.postprocessing.confidence_threshold
        logger.info(f"Confidence threshold {self.confidence_threshold}")

        dataset.purpose = DatasetPurpose.INFERENCE
        prediction_results, _ = self._infer_model(dataset, inference_parameters)

        self._add_predictions_to_dataset(
            prediction_results, dataset, self.confidence_threshold, process_saliency_maps, explain_predicted_classes
        )
        logger.info("Inference completed")
        return dataset

    @abstractmethod
    def _infer_model(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ):
        """Get inference results from dataset."""
        raise NotImplementedError

    def export(
        self,
        export_type: ExportType,
        output_model: ModelEntity,
        precision: ModelPrecision = ModelPrecision.FP32,
        dump_features: bool = True,
    ):
        """Export function of OTX Detection Task."""
        logger.info("Exporting the model")
        if export_type != ExportType.OPENVINO:
            raise RuntimeError(f"not supported export type {export_type}")
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.MO

        results = self._export_model(precision, dump_features)
        outputs = results.get("outputs")
        logger.debug(f"results of run_task = {outputs}")
        if outputs is None:
            raise RuntimeError(results.get("msg"))

        bin_file = outputs.get("bin")
        xml_file = outputs.get("xml")
        onnx_file = outputs.get("onnx")

        ir_extra_data = get_det_model_api_configuration(
            self._task_environment.label_schema, self._task_type, self.confidence_threshold
        )
        embed_ir_model_data(xml_file, ir_extra_data)

        if xml_file is None or bin_file is None or onnx_file is None:
            raise RuntimeError("invalid status of exporting. bin and xml or onnx should not be None")
        with open(bin_file, "rb") as f:
            output_model.set_data("openvino.bin", f.read())
        with open(xml_file, "rb") as f:
            output_model.set_data("openvino.xml", f.read())
        with open(onnx_file, "rb") as f:
            output_model.set_data("model.onnx", f.read())

        if self._hyperparams.tiling_parameters.enable_tile_classifier:
            tile_classifier = None
            for partition in outputs.get("partitioned", {}):
                if partition.get("tile_classifier"):
                    tile_classifier = partition.get("tile_classifier")
                    break
            if tile_classifier is None:
                raise RuntimeError("invalid status of exporting. tile_classifier should not be None")
            with open(tile_classifier["bin"], "rb") as f:
                output_model.set_data("tile_classifier.bin", f.read())
            with open(tile_classifier["xml"], "rb") as f:
                output_model.set_data("tile_classifier.xml", f.read())

        output_model.set_data(
            "confidence_threshold",
            np.array([self.confidence_threshold], dtype=np.float32).tobytes(),
        )
        output_model.set_data("config.json", config_to_bytes(self._hyperparams))
        output_model.precision = self._precision
        output_model.optimization_methods = self._optimization_methods
        output_model.has_xai = dump_features
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        logger.info("Exporting completed")

    @abstractmethod
    def _export_model(self, precision: ModelPrecision, dump_features: bool):
        """Main export function using training backend."""
        raise NotImplementedError

    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[ExplainParameters] = None,
    ) -> DatasetEntity:
        """Main explain function of OTX Task."""
        logger.info("explain()")

        update_progress_callback = default_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if explain_parameters is not None:
            update_progress_callback = explain_parameters.update_progress  # type: ignore
            process_saliency_maps = explain_parameters.process_saliency_maps
            explain_predicted_classes = explain_parameters.explain_predicted_classes

        self._time_monitor = InferenceProgressCallback(len(dataset), update_progress_callback)

        outputs = self._explain_model(dataset, explain_parameters)
        detections = outputs["detections"]
        explain_results = outputs["saliency_maps"]

        self._add_explanations_to_dataset(
            detections, explain_results, dataset, process_saliency_maps, explain_predicted_classes
        )
        logger.info("Explain completed")
        return dataset

    @abstractmethod
    def _explain_model(self, dataset: DatasetEntity, explain_parameters: Optional[ExplainParameters]):
        raise NotImplementedError

    def evaluate(
        self,
        output_resultset: ResultSetEntity,
        evaluation_metric: Optional[str] = None,
    ):
        """Evaluate function of OTX Detection Task."""
        logger.info("called evaluate()")
        if evaluation_metric is not None:
            logger.warning(
                f"Requested to use {evaluation_metric} metric, " "but parameter is ignored. Use F-measure instead."
            )
        metric = MetricsHelper.compute_f_measure(output_resultset)
        logger.info(f"F-measure after evaluation: {metric.f_measure.value}")
        output_resultset.performance = metric.get_performance()
        logger.info("Evaluation completed")

    def _add_predictions_to_dataset(
        self,
        prediction_results,
        dataset,
        confidence_threshold=0.0,
        process_saliency_maps=False,
        explain_predicted_classes=True,
    ):
        """Loop over dataset again to assign predictions. Convert from MMDetection format to OTX format."""
        for dataset_item, (all_results, feature_vector, saliency_map) in zip(dataset, prediction_results):
            shapes = self._get_shapes(all_results, dataset_item.width, dataset_item.height, confidence_threshold)
            dataset_item.append_annotations(shapes)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if saliency_map is not None:
                labels = self._labels.copy()
                if saliency_map.shape[0] == len(labels) + 1:
                    # Include the background as the last category
                    labels.append(LabelEntity("background", Domain.DETECTION))

                predicted_scored_labels = []
                for shape in shapes:
                    predicted_scored_labels += shape.get_labels()

                add_saliency_maps_to_dataset_item(
                    dataset_item=dataset_item,
                    saliency_map=saliency_map,
                    model=self._task_environment.model,
                    labels=labels,
                    predicted_scored_labels=predicted_scored_labels,
                    explain_predicted_classes=explain_predicted_classes,
                    process_saliency_maps=process_saliency_maps,
                )

    def _get_shapes(self, all_results, width, height, confidence_threshold):
        if self._task_type == TaskType.DETECTION:
            shapes = self._det_add_predictions_to_dataset(all_results, width, height, confidence_threshold)
        elif self._task_type in {
            TaskType.INSTANCE_SEGMENTATION,
            TaskType.ROTATED_DETECTION,
        }:
            shapes = self._ins_seg_add_predictions_to_dataset(all_results, width, height, confidence_threshold)
        else:
            raise RuntimeError(f"MPA results assignment not implemented for task: {self._task_type}")
        return shapes

    def _det_add_predictions_to_dataset(self, all_results, width, height, confidence_threshold):
        shapes = []
        for label_idx, detections in enumerate(all_results):
            for i in range(detections.shape[0]):
                probability = float(detections[i, 4])
                coords = detections[i, :4].astype(float).copy()
                coords /= np.array([width, height, width, height], dtype=float)
                coords = np.clip(coords, 0, 1)

                if probability < confidence_threshold:
                    continue

                assigned_label = [ScoredLabel(self._labels[label_idx], probability=probability)]
                if coords[3] - coords[1] <= 0 or coords[2] - coords[0] <= 0:
                    continue

                shapes.append(
                    Annotation(
                        Rectangle(x1=coords[0], y1=coords[1], x2=coords[2], y2=coords[3]),
                        labels=assigned_label,
                    )
                )
        return shapes

    def _ins_seg_add_predictions_to_dataset(self, all_results, width, height, confidence_threshold):
        shapes = []
        for label_idx, (boxes, masks) in enumerate(zip(*all_results)):
            for mask, probability in zip(masks, boxes[:, 4]):
                if isinstance(mask, dict):
                    mask = mask_util.decode(mask)
                mask = mask.astype(np.uint8)
                probability = float(probability)
                contours, hierarchies = cv2.findContours(mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
                if hierarchies is None:
                    continue
                for contour, hierarchy in zip(contours, hierarchies[0]):
                    if hierarchy[3] != -1:
                        continue
                    if len(contour) <= 2 or probability < confidence_threshold:
                        continue
                    if self._task_type == TaskType.INSTANCE_SEGMENTATION:
                        points = [Point(x=point[0][0] / width, y=point[0][1] / height) for point in contour]
                    else:
                        box_points = cv2.boxPoints(cv2.minAreaRect(contour))
                        points = [Point(x=point[0] / width, y=point[1] / height) for point in box_points]
                    labels = [ScoredLabel(self._labels[label_idx], probability=probability)]
                    polygon = Polygon(points=points)
                    if cv2.contourArea(contour) > 0 and polygon.get_area() > 1e-12:
                        shapes.append(Annotation(polygon, labels=labels, id=ID(f"{label_idx:08}")))
        return shapes

    def _add_explanations_to_dataset(
        self, detections, explain_results, dataset, process_saliency_maps, explain_predicted_classes
    ):
        """Add saliency map to the dataset."""
        for dataset_item, detection, saliency_map in zip(dataset, detections, explain_results):
            labels = self._labels.copy()
            if saliency_map.shape[0] == len(labels) + 1:
                # Include the background as the last category
                labels.append(LabelEntity("background", Domain.DETECTION))

            shapes = self._get_shapes(detection, dataset_item.width, dataset_item.height, 0.4)
            predicted_scored_labels = []
            for shape in shapes:
                predicted_scored_labels += shape.get_labels()

            add_saliency_maps_to_dataset_item(
                dataset_item=dataset_item,
                saliency_map=saliency_map,
                model=self._task_environment.model,
                labels=labels,
                predicted_scored_labels=predicted_scored_labels,
                explain_predicted_classes=explain_predicted_classes,
                process_saliency_maps=process_saliency_maps,
            )

    @staticmethod
    def _generate_training_metrics(learning_curves, scores) -> Iterable[MetricsGroup[Any, Any]]:
        """Get Training metrics (epochs & scores).

        Parses the mmdetection logs to get metrics from the latest training run
        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []

        # Learning curves.
        for key, curve in learning_curves.items():
            len_x, len_y = len(curve.x), len(curve.y)
            if len_x != len_y:
                logger.warning(f"Learning curve {key} has inconsistent number of coordinates ({len_x} vs {len_y}.")
                len_x = min(len_x, len_y)
                curve.x = curve.x[:len_x]
                curve.y = curve.y[:len_x]
            metric_curve = CurveMetric(
                xs=np.nan_to_num(curve.x).tolist(),
                ys=np.nan_to_num(curve.y).tolist(),
                name=key,
            )
            visualization_info = LineChartInfo(name=key, x_axis_label="Epoch", y_axis_label=key)
            output.append(LineMetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))

        # Final mAP value on the validation set.
        output.append(
            BarMetricsGroup(
                metrics=[ScoreMetric(value=scores, name="mAP")],
                visualization_info=BarChartInfo("Validation score", visualization_type=VisualizationType.RADIAL_BAR),
            )
        )

        return output

    def save_model(self, output_model: ModelEntity):
        """Save best model weights in DetectionTrainTask."""
        if is_multigpu_child_process():
            return

        logger.info("called save_model")
        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}
        model_ckpt = torch.load(self._model_ckpt)
        modelinfo = {
            "model": model_ckpt,
            "config": hyperparams_str,
            "labels": labels,
            "confidence_threshold": self.confidence_threshold,
            "VERSION": 1,
        }

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        output_model.precision = self._precision



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import openvino.runtime as ov

from otx.core.ov.ops.arithmetics import MultiplyV1
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestOperation:
    @e2e_pytest_unit
    def test(self):
        data = ov.opset10.parameter([3, 1, 2], ov.Type.f32)
        mul_constant = ov.opset10.constant([1.5], ov.Type.f32)
        mul = ov.opset10.multiply(data, mul_constant)
        op = MultiplyV1.from_ov(mul)
        assert isinstance(op, MultiplyV1)
        assert op.type == MultiplyV1.TYPE
        assert op.version == MultiplyV1.VERSION
        assert op.name == op._name
        assert op.attrs == op._attrs
        assert op.shape == op.attrs.shape
        assert isinstance(repr(op), str)



"""Collections of hooks for common OTX algorithms."""

# Copyright (C) 2021-2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import math

from mmcv.runner import BaseRunner
from mmcv.runner.hooks import HOOKS, Hook

from otx.algorithms.common.utils.logger import get_logger
from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback

logger = get_logger()


@HOOKS.register_module()
class OTXProgressHook(Hook):
    """OTXProgressHook for getting progress."""

    def __init__(self, time_monitor: TimeMonitorCallback, verbose: bool = False):
        super().__init__()
        self.time_monitor = time_monitor
        self.verbose = verbose
        self.print_threshold = 1

    def before_run(self, runner: BaseRunner):
        """Called before_run in OTXProgressHook."""
        total_epochs = runner.max_epochs if runner.max_epochs is not None else 1
        self.time_monitor.total_epochs = total_epochs
        self.time_monitor.train_steps = runner.max_iters // total_epochs if total_epochs else 1
        self.time_monitor.steps_per_epoch = self.time_monitor.train_steps + self.time_monitor.val_steps
        self.time_monitor.total_steps = max(math.ceil(self.time_monitor.steps_per_epoch * total_epochs), 1)
        self.time_monitor.current_step = 0
        self.time_monitor.current_epoch = 0
        self.time_monitor.on_train_begin()

    def before_epoch(self, runner: BaseRunner):
        """Called before_epoch in OTXProgressHook."""
        self.time_monitor.on_epoch_begin(runner.epoch)

    def after_epoch(self, runner: BaseRunner):
        """Called after_epoch in OTXProgressHook."""
        # put some runner's training status to use on the other hooks
        runner.log_buffer.output["current_iters"] = runner.iter
        self.time_monitor.on_epoch_end(runner.epoch, runner.log_buffer.output)

    def before_iter(self, runner: BaseRunner):
        """Called before_iter in OTXProgressHook."""
        self.time_monitor.on_train_batch_begin(1)

    def after_iter(self, runner: BaseRunner):
        """Called after_iter in OTXProgressHook."""
        # put some runner's training status to use on the other hooks
        runner.log_buffer.output["current_iters"] = runner.iter
        self.time_monitor.on_train_batch_end(1)
        if self.verbose:
            progress = self.progress
            if progress >= self.print_threshold:
                logger.warning(f"training progress {progress:.0f}%")
                self.print_threshold = (progress + 10) // 10 * 10

    def before_val_iter(self, runner: BaseRunner):
        """Called before_val_iter in OTXProgressHook."""
        self.time_monitor.on_test_batch_begin(1, logger)

    def after_val_iter(self, runner: BaseRunner):
        """Called after_val_iter in OTXProgressHook."""
        self.time_monitor.on_test_batch_end(1, logger)

    def after_run(self, runner: BaseRunner):
        """Called after_run in OTXProgressHook."""
        self.time_monitor.on_train_end(1)
        if self.time_monitor.update_progress_callback:
            self.time_monitor.update_progress_callback(int(self.time_monitor.get_progress()))

    @property
    def progress(self):
        """Getting Progress from time monitor."""
        return self.time_monitor.get_progress()



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.core.ov.graph.graph import Graph
from otx.core.ov.graph.utils import (
    get_constant_input_nodes,
    handle_paired_batchnorm,
    handle_reshape,
)
from otx.core.ov.utils import load_ov_model
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def get_graph(name="mobilenet-v2-pytorch"):
    model = load_ov_model(f"omz://{name}")
    return Graph.from_ov(model)


@e2e_pytest_unit
def test_get_constant_input_nodes():
    graph = get_graph()
    node = graph.get_nodes_by_types(["Add"])[0]
    nodes = get_constant_input_nodes(graph, node)
    assert all([node.type == "Constant" for node in nodes])


@e2e_pytest_unit
def test_handle_merging_into_batchnorm():
    # TODO:
    pass
    #  graph = get_graph()
    #  n_nodes = len(graph)
    #  handle_merging_into_batchnorm(graph)
    #
    #  assert graph.get_nodes_by_types(["BatchNormInference"])
    #  assert n_nodes >= len(graph)


@e2e_pytest_unit
def test_handle_paired_batchnorm():
    graph = get_graph()
    handle_paired_batchnorm(graph)
    n_nodes = len(graph)
    assert graph.get_nodes_by_types(["BatchNormInference"])

    graph = get_graph()
    handle_paired_batchnorm(graph, replace=True)
    assert graph.get_nodes_by_types(["BatchNormInference"])

    assert n_nodes > len(graph)


@e2e_pytest_unit
def test_handle_reshape():
    graph = get_graph("dla-34")
    handle_reshape(graph)



from os import path as osp
from tempfile import TemporaryDirectory

import pytest

from otx.cli.utils.nncf import get_number_of_fakequantizers_in_xml, is_checkpoint_nncf
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_is_checkpoint_nncf_meta_exist(mocker):
    mock_state = {"meta": {"nncf_enable_compression": "fake"}}
    mocker.patch("torch.load").return_value = mock_state

    assert is_checkpoint_nncf("fake")

    # state["meta"]["nncf_enable_compression"]
    # state["nncf_enable_compression"]


@e2e_pytest_unit
def test_is_checkpoint_nncf_metainfo_exist(mocker):
    mock_state = {"nncf_metainfo": "fake"}
    mocker.patch("torch.load").return_value = mock_state

    assert is_checkpoint_nncf("fake")


@e2e_pytest_unit
def test_is_not_checkpoint_nncf_meta_not_exist(mocker):
    mock_state = {"fake": "fake"}
    mocker.patch("torch.load").return_value = mock_state

    assert not is_checkpoint_nncf("fake")


@e2e_pytest_unit
@pytest.mark.parametrize("number_of_fakequantizers_in_xml", [0, 10, 100])
def test_get_number_of_fakequantizers_in_xml(number_of_fakequantizers_in_xml):
    with TemporaryDirectory() as tmp_dir:
        path_to_xml = osp.join(tmp_dir, "fake.xml")
        with open(path_to_xml, "w") as f:
            for _ in range(number_of_fakequantizers_in_xml):
                f.write('type="FakeQuantize"\n')

        assert get_number_of_fakequantizers_in_xml(path_to_xml) == number_of_fakequantizers_in_xml



"""Redunction-related modules for otx.core.ov.ops."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field

import torch

from otx.core.ov.ops.builder import OPS
from otx.core.ov.ops.op import Attribute, Operation


@dataclass
class ReduceMeanV1Attribute(Attribute):
    """ReduceMeanV1Attribute class."""

    keep_dims: bool = field(default=False)


@OPS.register()
class ReduceMeanV1(Operation[ReduceMeanV1Attribute]):
    """ReduceMeanV1 class."""

    TYPE = "ReduceMean"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReduceMeanV1Attribute

    def forward(self, inputs, axes):
        """ReduceMeanV1's forward function."""
        if isinstance(axes, torch.Tensor):
            axes = axes.tolist()
        if not axes:
            return inputs

        if not isinstance(axes, (list, tuple)):
            axes = [axes]

        return torch.mean(input=inputs, dim=axes, keepdim=self.attrs.keep_dims)


@dataclass
class ReduceProdV1Attribute(Attribute):
    """ReduceMeanV1Attribute class."""

    keep_dims: bool = field(default=False)


@OPS.register()
class ReduceProdV1(Operation[ReduceProdV1Attribute]):
    """ReduceMeanV1Attribute class."""

    TYPE = "ReduceProd"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReduceProdV1Attribute

    def forward(self, inputs, axes):
        """ReduceMeanV1Attribute's forward function."""
        if isinstance(axes, torch.Tensor):
            axes = axes.tolist()
        if not axes:
            return inputs

        if not isinstance(axes, (list, tuple)):
            axes = [axes]

        output = inputs
        for axe in axes:
            output = torch.prod(input=output, dim=axe, keepdim=True)
        if not self.attrs.keep_dims:
            output = torch.squeeze(output)

        return output


@dataclass
class ReduceMinV1Attribute(Attribute):
    """ReduceMinV1Attribute class."""

    keep_dims: bool = field(default=False)


@OPS.register()
class ReduceMinV1(Operation[ReduceMinV1Attribute]):
    """ReduceMinV1 class."""

    TYPE = "ReduceMin"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReduceMinV1Attribute

    def forward(self, inputs, axes):
        """ReduceMinV1's forward function."""
        if isinstance(axes, torch.Tensor):
            axes = axes.tolist()
        if not axes:
            return inputs

        if not isinstance(axes, (list, tuple)):
            axes = [axes]

        output = inputs
        for axe in axes:
            output = torch.min(input=output, dim=axe, keepdim=True)[0]
        if not self.attrs.keep_dims:
            output = torch.squeeze(output)

        return output


@dataclass
class ReduceSumV1Attribute(Attribute):
    """ReduceSumV1Attribute class."""

    keep_dims: bool = field(default=False)


@OPS.register()
class ReduceSumV1(Operation[ReduceSumV1Attribute]):
    """ReduceSumV1 class."""

    TYPE = "ReduceSum"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReduceSumV1Attribute

    def forward(self, inputs, axes):
        """ReduceSumV1's forward function."""
        if isinstance(axes, torch.Tensor):
            axes = axes.tolist()
        if not axes:
            return inputs

        return torch.sum(input=inputs, dim=axes, keepdim=self.attrs.keep_dims)



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from copy import deepcopy
import os
import stat
import tempfile
import time

import numpy as np
import pytest
from datumaro.components.annotation import Label, Polygon, Bbox, Mask
from datumaro.components.dataset import Dataset
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.media import Image

from otx.core.data.caching.storage_cache import init_arrow_cache


@pytest.fixture
def fxt_datumaro_dataset():
    items = []
    for i in range(64):
        media = Image.from_numpy(data=np.random.randint(0, 255, (5, 5, 3)), ext=".png")

        items.append(
            DatasetItem(
                id=i,
                subset="test",
                media=media,
                annotations=[
                    # annotations used in OTX
                    Label(np.random.randint(0, 3)),
                    Bbox(*np.random.randint(0, 5, 4)),
                    Polygon(Bbox(*np.random.randint(0, 5, 4)).as_polygon()),
                    Mask(np.random.randint(0, 2, (5, 5))),
                ],
            )
        )

    source_dataset = Dataset.from_iterable(
        items,
        categories=["label"],
        media_type=Image,
    )

    return source_dataset


def compare_dataset(source_dataset, target_dataset, compare_media=True):
    properties = ["id", "subset", "annotations", "attributes"]
    if compare_media:
        properties.append("media")
    for item_s, item_t in zip(source_dataset, target_dataset):
        for property in properties:
            assert getattr(item_s, property) == getattr(item_t, property)


class TestStorageCache:
    @pytest.mark.parametrize(
        ["scheme", "compare_media"],
        [
            pytest.param(
                "NONE",
                True,
                id="test_none_scheme",
            ),
            pytest.param(
                "AS-IS",
                True,
                id="test_as_is_scheme",
            ),
            pytest.param(
                "PNG",
                True,
                id="test_png_scheme",
            ),
            pytest.param(
                "TIFF",
                True,
                id="test_tiff_scheme",
            ),
            pytest.param(
                "JPEG/95",
                False,
                id="test_jpeg_95_scheme",
            ),
            pytest.param(
                "JPEG/75",
                False,
                id="test_jpeg_75_scheme",
            ),
        ],
    )
    def test_is_identical(self, scheme, fxt_datumaro_dataset, compare_media):
        with tempfile.TemporaryDirectory() as tempdir:
            source_dataset = fxt_datumaro_dataset
            cached_dataset = init_arrow_cache(source_dataset, scheme=scheme, cache_dir=tempdir)
            compare_dataset(source_dataset, cached_dataset, compare_media)

    def test_cache_hit(self, fxt_datumaro_dataset):
        with tempfile.TemporaryDirectory() as tempdir:
            cached_dataset = init_arrow_cache(deepcopy(fxt_datumaro_dataset), scheme="AS-IS", cache_dir=tempdir)

            mapping = {}
            for file in os.listdir(cached_dataset.data_path):
                mapping[file] = os.stat(os.path.join(cached_dataset.data_path, file))[stat.ST_MTIME]

            cached_dataset = init_arrow_cache(deepcopy(fxt_datumaro_dataset), scheme="AS-IS", cache_dir=tempdir)

            for file in os.listdir(cached_dataset.data_path):
                assert mapping[file] == os.stat(os.path.join(cached_dataset.data_path, file))[stat.ST_MTIME]

    def test_no_cache_hit(self, fxt_datumaro_dataset):
        with tempfile.TemporaryDirectory() as tempdir:
            cached_dataset = init_arrow_cache(deepcopy(fxt_datumaro_dataset), scheme="AS-IS", cache_dir=tempdir)

            mapping = {}
            for file in os.listdir(cached_dataset.data_path):
                mapping[file] = os.stat(os.path.join(cached_dataset.data_path, file))[stat.ST_MTIME]

            # sleep 1 second to invalidate cache
            time.sleep(1)

            cached_dataset = init_arrow_cache(
                deepcopy(fxt_datumaro_dataset), scheme="AS-IS", cache_dir=tempdir, force=True
            )

            for file in os.listdir(cached_dataset.data_path):
                assert mapping[file] != os.stat(os.path.join(cached_dataset.data_path, file))[stat.ST_MTIME]



"""Detection Saliency Map Hook."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from typing import List, Tuple, Union

import torch
import torch.nn.functional as F

from otx.algorithms.common.adapters.mmcv.hooks.recording_forward_hook import (
    BaseRecordingForwardHook,
)
from otx.algorithms.detection.adapters.mmdet.models.heads.custom_atss_head import (
    CustomATSSHead,
)
from otx.algorithms.detection.adapters.mmdet.models.heads.custom_ssd_head import (
    CustomSSDHead,
)
from otx.algorithms.detection.adapters.mmdet.models.heads.custom_vfnet_head import (
    CustomVFNetHead,
)
from otx.algorithms.detection.adapters.mmdet.models.heads.custom_yolox_head import (
    CustomYOLOXHead,
)

# pylint: disable=too-many-locals


class DetClassProbabilityMapHook(BaseRecordingForwardHook):
    """Saliency map hook for object detection models."""

    def __init__(self, module: torch.nn.Module) -> None:
        super().__init__(module)
        self._neck = module.neck if module.with_neck else None
        self._bbox_head = module.bbox_head
        self._num_cls_out_channels = module.bbox_head.cls_out_channels  # SSD-like heads also have background class
        if hasattr(module.bbox_head, "anchor_generator"):
            self._num_anchors = module.bbox_head.anchor_generator.num_base_anchors
        else:
            self._num_anchors = [1] * 10

    def func(
        self,
        feature_map: Union[torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor]],
        _: int = -1,
        cls_scores_provided: bool = False,
    ) -> torch.Tensor:
        """Generate the saliency map from raw classification head output, then normalizing to (0, 255).

        :param x: Feature maps from backbone/FPN or classification scores from cls_head
        :param cls_scores_provided: If True - use 'x' as is, otherwise forward 'x' through the classification head
        :return: Class-wise Saliency Maps. One saliency map per each class - [batch, class_id, H, W]
        """
        if cls_scores_provided:
            cls_scores = feature_map
        else:
            cls_scores = self._get_cls_scores_from_feature_map(feature_map)

        batch_size, _, height, width = cls_scores[-1].size()
        saliency_maps = torch.empty(batch_size, self._num_cls_out_channels, height, width)
        for batch_idx in range(batch_size):
            cls_scores_anchorless = []
            for scale_idx, cls_scores_per_scale in enumerate(cls_scores):
                cls_scores_anchor_grouped = cls_scores_per_scale[batch_idx].reshape(
                    self._num_anchors[scale_idx], (self._num_cls_out_channels), *cls_scores_per_scale.shape[-2:]
                )
                cls_scores_out, _ = cls_scores_anchor_grouped.max(dim=0)
                cls_scores_anchorless.append(cls_scores_out.unsqueeze(0))
            cls_scores_anchorless_resized = []
            for cls_scores_anchorless_per_level in cls_scores_anchorless:
                cls_scores_anchorless_resized.append(
                    F.interpolate(cls_scores_anchorless_per_level, (height, width), mode="bilinear")
                )
            saliency_maps[batch_idx] = torch.cat(cls_scores_anchorless_resized, dim=0).mean(dim=0)

        saliency_maps = saliency_maps.reshape((batch_size, self._num_cls_out_channels, -1))
        max_values, _ = torch.max(saliency_maps, -1)
        min_values, _ = torch.min(saliency_maps, -1)
        saliency_maps = 255 * (saliency_maps - min_values[:, :, None]) / (max_values - min_values + 1e-12)[:, :, None]
        saliency_maps = saliency_maps.reshape((batch_size, self._num_cls_out_channels, height, width))
        saliency_maps = saliency_maps.to(torch.uint8)
        return saliency_maps

    def _get_cls_scores_from_feature_map(self, x: torch.Tensor) -> List:
        """Forward features through the classification head of the detector."""
        with torch.no_grad():
            if self._neck is not None:
                x = self._neck(x)

            if isinstance(self._bbox_head, CustomSSDHead):
                cls_scores = []
                for feat, cls_conv in zip(x, self._bbox_head.cls_convs):
                    cls_scores.append(cls_conv(feat))
            elif isinstance(self._bbox_head, CustomATSSHead):
                cls_scores = []
                for cls_feat in x:
                    for cls_conv in self._bbox_head.cls_convs:
                        cls_feat = cls_conv(cls_feat)
                    cls_score = self._bbox_head.atss_cls(cls_feat)
                    cls_scores.append(cls_score)
            elif isinstance(self._bbox_head, CustomVFNetHead):
                # Not clear how to separate cls_scores from bbox_preds
                cls_scores, _, _ = self._bbox_head(x)
            elif isinstance(self._bbox_head, CustomYOLOXHead):

                def forward_single(x, cls_convs, conv_cls):
                    """Forward feature of a single scale level."""
                    cls_feat = cls_convs(x)
                    cls_score = conv_cls(cls_feat)
                    return cls_score

                map_results = map(
                    forward_single, x, self._bbox_head.multi_level_cls_convs, self._bbox_head.multi_level_conv_cls
                )
                cls_scores = list(map_results)
            else:
                raise NotImplementedError(
                    "Not supported detection head provided. "
                    "DetClassProbabilityMap supports only the following single stage detectors: "
                    "YOLOXHead, ATSSHead, SSDHead, VFNetHead."
                )
        return cls_scores



"""OTX Algorithms - Classification Dataset."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .otx_datasets import (
    OTXClsDataset,
    OTXHierarchicalClsDataset,
    OTXMultilabelClsDataset,
    SelfSLDataset,
)
from .pipelines.transforms import (
    AugMixAugment,
    OTXRandAugment,
    PILToTensor,
    RandomRotate,
    TensorNormalize,
    TwoCropTransform,
)

__all__ = [
    "AugMixAugment",
    "PILToTensor",
    "TensorNormalize",
    "RandomRotate",
    "OTXRandAugment",
    "TwoCropTransform",
    "OTXClsDataset",
    "OTXMultilabelClsDataset",
    "OTXHierarchicalClsDataset",
    "SelfSLDataset",
]
